#  Copyright (c) 2022 by Amplo.

import warnings
from pathlib import Path
from typing import Union

import faiss
import numpy as np
import pandas as pd

from amplo.automl import DataProcessor
from amplo.automl.feature_processing import FeatureProcessor
from amplo.utils.data import check_dataframe_quality, check_pearson_correlation
from amplo.utils.io import merge_logs
from amplo.utils.logging import logger


class IntervalAnalyser:

    noise = -1

    def __init__(
        self,
        target: str = None,
        norm: str = "euclidean",  # TODO: implement functionality
        min_length: int = 1000,
        n_neighbors: int = None,
        n_trees: int = 10,
        verbose: int = 1,
    ):
        """
        Interval Analyser for Log file classification. Has two purposes:
        - Remove healthy data in longer, faulty logs
        - Remove redundant data in large datasets

        Uses Facebook's FAISS for K-Nearest Neighbors approximation.

        ** IMPORTANT **
        To use this interval analyser, make sure that your logs are located in a
        folder of their label, with one parent folder with all labels, e.g.:
        +-- Parent Folder
        |   +-- Label_1
        |       +-- Log_1.*
        |       +-- Log_2.*
        |   +-- Label_2
        |       +-- Log_3.*

        Parameters
        ----------
        target : str
            Target column name
        norm : str
            Optimization metric for K-Nearest Neighbors
            NOTE: This option has no effect, yet!
        min_length : int
            Minimum length to cut off, everything shorter is left untouched
        n_neighbors : int, optional
            Quantity of neighbors, default to 3 * log length
        n_trees : int
            Quantity of trees
        """
        # Test
        self.available_norms = ["euclidean", "manhattan", "angular", "hamming", "dot"]
        if norm not in self.available_norms:
            raise ValueError(f"Unknown norm, pick from {self.available_norms}")
        # Parameters
        self.target = target or "labels"
        self.min_length = min_length
        self.norm = norm
        self.n_trees = n_trees
        self.n_neighbors = n_neighbors
        self.verbose = verbose

        # Initializers
        self._labels = None
        self._features = None
        self._metadata = None  # optional metadata from `merge_logs`
        # self._mins = pd.DataFrame(index=[0])
        # self._maxs = pd.DataFrame(index=[0])
        self._engine = None
        self._distributions = None
        self._noise_indices = None
        self.avg_samples = None
        self.n_samples = None
        self.n_columns = None
        self.n_files = None
        self.n_folders = None

        # Flags
        self.is_fitted = False

    @property
    def orig_data(self):
        """
        Returns
        -------
        pd.DataFrame
            Original data containing both, feature and label columns
        """
        return pd.concat([self._features, self._labels], axis=1)

    @property
    def data_with_noise(self):
        """
        Returns
        -------
        pd.DataFrame
            Original data where noise is labelled as ``self.noise``

        Notes
        -----
        Depending on the dtype of the noise attribute, the dtype of the labels' column
        will be affected.
        """
        data = self.orig_data
        data.loc[self._noise_indices, self.target] = self.noise
        return data

    @property
    def data_without_noise(self):
        """
        Returns
        -------
        pd.DataFrame
            Original data but without noise rows
        """
        return self.orig_data.drop(self._noise_indices)

    def fit_transform(
        self, data_or_path: Union[pd.DataFrame, str, Path], labels: pd.Series = None
    ):
        """
        Function that runs the K-Nearest Neighbors and returns a dataframe with the
        sensitivities.

        Parameters
        ----------
        data_or_path : pd.DataFrame or str or Path
            Multi-indexed dataset containing feature (and target) columns
            or path to folder that is correctly structured (see Notes)
        labels : pd.Series, optional
            Multi-indexed dataset containing target columns.

        Notes
        -----
        In case you provide a path-like argument to ``features_or_path``,
        make sure that each protocol is located in a sub folder whose name represents
        the respective label.

        A directory structure example:
            |   ``path_to_folder``
            |   ``├─ Label_1``
            |   ``│   ├─ Log_1.*``
            |   ``│   └─ Log_2.*``
            |   ``├─ Label_2``
            |   ``│   └─ Log_3.*``
            |   ``└─ ...``

        Returns
        -------
        pd.DataFrame
            Estimation of correlation strength
        """

        # Parse data
        features, labels = self._parse_data(data_or_path, labels)

        # Set up Annoy Engine (only now that n_keys is known)
        self._build_engine(features)

        # Make distribution
        self._make_distribution(features, labels)

        # Filter dataset
        self._set_noise_indices()

        # Set flags
        self.is_fitted = True

        return self.data_with_noise

    def _parse_data(
        self, data_or_path: Union[pd.DataFrame, str, Path], labels: pd.Series = None
    ):
        """
        Parse data for interval analysis and store it internally

        Parameters
        ----------
        data_or_path : pd.DataFrame or str or Path
            Multi-indexed dataset containing feature (and target) columns
            or path to folder that is correctly structured (see Notes)
        labels : pd.Series, optional
            Multi-indexed dataset containing target columns
        """
        # Verbose
        if self.verbose > 0:
            logger.info("Parsing data for interval analyser.")

        # Parse data
        if isinstance(data_or_path, (str, Path)):
            # Read from path
            features, metadata = merge_logs(data_or_path, self.target)
            labels = features.pop(self.target)
            # Store metadata in self
            self._metadata = metadata
        else:
            if not isinstance(data_or_path, pd.DataFrame):
                raise ValueError("Invalid data_or_path type.")
            # Check if target in `data_or_path`
            if self.target in data_or_path.columns:
                features = data_or_path.drop(self.target, axis=1)
                labels = data_or_path.loc[:, self.target]
            else:
                if not isinstance(labels, pd.Series):
                    raise ValueError("Invalid labels data type.")
                features = data_or_path

        # Tests
        if list(features.index.names) != ["log", "index"]:
            raise ValueError("Invalid multi-indexed data detected.")
        if not all(features.index == labels.index):
            raise ValueError(
                "Indices mismatch: Features and labels cannot be concatenated."
            )
        if len(features) != len(labels):
            raise ValueError(
                "Length mismatch: Features and labels cannot be concatenated."
            )
        if self.target in features.columns:
            raise ValueError("Target column is present in feature data.")

        # Check data quality
        if not check_dataframe_quality(features):
            # Warn & clean
            warnings.warn("Data quality is insufficient, starting DataProcessor.")
            features = DataProcessor().fit_transform(features)

        # Check collinearity
        if not check_pearson_correlation(features, labels):
            warnings.warn("Data is correlated, starting FeatureProcessor.")
            # Remove collinear features and pick Random Forest Increment
            fp = FeatureProcessor(mode="classification", extract_features=False)
            fp.fit(features, labels)
            features = fp.transform(features, feature_set="rf_increment")

        # Set name of labels
        if self.target != labels.name:
            warnings.warn(
                "Expected target name does not match the actual name. "
                f"Name will be fixed from {labels.name} to {self.target}."
            )
            labels.name = self.target

        # Remove datetime columns
        _date_cols = [
            col
            for col in features.columns
            if pd.api.types.is_datetime64_any_dtype(features[col])
        ]
        if len(_date_cols) != 0:
            features.drop(_date_cols, axis=1, inplace=True)

        # Normalize
        self._mins, self._maxs = features.min(), features.max()
        features = (features - features.min()) / (features.max() - features.min())

        # Change float64 to float32
        #  See: https://github.com/facebookresearch/faiss/issues/461
        for col in features.select_dtypes(include=["float64"]).columns:
            features[col] = features[col].astype("float32")

        # Store data in self
        self._features = features
        self._labels = labels

        # Set sizes
        self.n_folders = labels.nunique()
        self.n_files = features.index.get_level_values(0).nunique()
        self.n_samples, self.n_columns = features.shape
        if self.n_neighbors is None:
            self.n_neighbors = min(3 * self.n_samples // self.n_files, 5000)

        return features, labels

    def _build_engine(self, features: pd.DataFrame):
        """
        Builds the ANNOY engine.

        Parameters
        ----------
        features : pd.DataFrame
            Multi-indexed dataset containing feature columns
        """
        # Create engine
        if self.verbose > 0:
            logger.info("Building interval analyser engine.")
        engine = faiss.IndexFlatL2(self.n_columns)

        # Add the data to ANNOY
        engine.add(np.ascontiguousarray(features.values))  # noqa

        # Set class attribute
        self._engine = engine

    def _make_distribution(self, features: pd.DataFrame, labels: pd.Series):
        """
        Given a build K-Nearest Neighbors, returns the label distribution

        Parameters
        ----------
        features : pd.DataFrame
            Multi-indexed dataset containing feature columns
        labels : pd.Series
            Multi-indexed dataset containing target columns
        """
        if self.verbose > 0:
            logger.info("Calculating interval within-class distributions.")

        # Search nearest neighbors for all samples - has to be iterative for large files
        distribution = []
        for i, row in features.iterrows():
            _, neighbors = self._engine.search(
                np.ascontiguousarray(row.values.reshape((1, -1))), int(self.n_neighbors)
            )
            match_mask = labels.iloc[neighbors.reshape(-1)] == labels.loc[i]
            distribution.append(pd.Series(match_mask).sum() / self.n_neighbors)

        # Parse into list of lists
        self._distributions = pd.Series(distribution, index=self._labels.index)

    def _set_noise_indices(self):
        """
        This function selects samples given the calculated distributions. It only
        removes samples from logs which are longer (> min_length), and only the samples
        with lower in-class neighbors.

        One could come up with a fancier logic, using the total dataset samples, the
        class-balance & sample redundancy.
        """
        # Verbose
        if self.verbose > 0:
            logger.info("Creating filtered dataset")

        # Get in-class means and number of samples for each label
        # label_means = self._get_label_means(labels, self._distributions)
        # label_samples = labels.value_counts()

        # Initialize
        data_index = self._features.index
        noise_indices = []

        # Iterate through labels and see if we should remove values
        for file_id in range(self.n_files):

            # Check length and continue if short
            if sum(data_index.get_level_values(0) == file_id) < self.min_length:
                continue

            # Check distribution and find cut-off
            dist = self._distributions[file_id]
            ind_remove_label = [(file_id, j) for j in np.where(dist < dist.mean())[0]]
            # Extend list to keep track
            noise_indices.extend(ind_remove_label)

            # Verbose
            if len(noise_indices) > 0 and self.verbose > 1:
                filename = (
                    self._metadata[file_id]["file"]
                    if isinstance(self._metadata, dict)
                    else None
                )
                logger.info(
                    f"Removing {len(ind_remove_label)} samples from "
                    f"`{filename or file_id}`"
                )

        # Set noise indices
        self._noise_indices = noise_indices

    # @staticmethod
    # def _get_label_means(labels, dists):
    #     # Init
    #     label_means = {k: [] for k in labels.unique()}
    #
    #     # Append distributions
    #     for i, d in enumerate(dists):
    #         # Skip failed reads
    #         if i not in labels:
    #             continue
    #
    #         # Extend distribution
    #         label_means[labels[(i, 0)]].extend(d)
    #
    #     # Return means
    #     return {k: np.mean(v) for k, v in label_means.items()}

    # --- Deprecation Properties ---

    @property
    def samples(self):
        warnings.warn(
            DeprecationWarning(
                "This pseudo-attribute will be removed in a future version. "
                "Consider using `n_samples` instead."
            )
        )
        return self.n_samples

    @property
    def n_keys(self):
        warnings.warn(
            DeprecationWarning(
                "This pseudo-attribute will be removed in a future version. "
                "Consider using `n_columns` instead."
            )
        )
        return self.n_columns



#  Copyright (c) 2022 by Amplo.

import time
import warnings

import numpy as np
import pytest
from sklearn.datasets import make_classification, make_regression

from amplo import Pipeline
from amplo.classification import CatBoostClassifier
from amplo.observation._base import ProductionWarning
from amplo.observation.model import ModelObserver
from amplo.regression import CatBoostRegressor
from tests import OverfitPredictor, RandomPredictor


@pytest.fixture
def make_one_to_one_data(mode):
    size = 100
    if mode == "classification":
        linear_col = np.random.choice([0, 1, 2], size)
    elif mode == "regression":
        linear_col = np.random.uniform(0.0, 1.0, size)
    else:
        raise ValueError("Invalid mode")
    x = linear_col.reshape(-1, 1)
    y = linear_col.reshape(-1)
    yield x, y


class DelayedRandomPredictor(RandomPredictor):
    def __init__(self, delay: float, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.delay = delay

    def predict(self, x):
        time.sleep(self.delay)
        return super().predict(x)

    def predict_proba(self, x):
        time.sleep(self.delay)
        return super().predict_proba(x)


class TestModelObserver:
    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_check_model_size(self, mode, make_x_y):
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline._read_data(*make_x_y)
        pipeline._mode_detector()
        pipeline.best_model = RandomPredictor(mode=mode)

        # Observe
        obs = ModelObserver(pipeline=pipeline)
        with warnings.catch_warnings(record=True) as record:
            obs.check_model_size()
        assert not any(
            isinstance(r.message, ProductionWarning) for r in record
        ), "An unnecessary warning was raised while checking model size."

        # Add big chunk of data to model to make it big
        pipeline.best_model.some_data = np.random.normal(size=3_000_000)
        with pytest.warns(ProductionWarning):
            obs.check_model_size()

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_check_better_than_linear(self, mode, make_one_to_one_data):
        x, y = make_one_to_one_data

        # Make pipeline and simulate fit
        pipeline = Pipeline(n_grid_searches=0)
        pipeline._read_data(x, y)
        pipeline._mode_detector()
        pipeline.best_model = RandomPredictor(mode=mode)

        # Observe
        obs = ModelObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning):
            obs.check_better_than_linear()

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_check_noise_invariance(self, mode, make_one_to_one_data):
        x, y = make_one_to_one_data

        # Make pipeline and simulate fit
        pipeline = Pipeline(n_grid_searches=0)
        pipeline._read_data(x, y)
        pipeline._mode_detector()
        pipeline.best_model = OverfitPredictor(mode=mode)

        # Observe
        obs = ModelObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning):
            obs.check_noise_invariance()

        # Should not trigger normally
        pipeline.best_model = RandomPredictor(mode=mode)
        obs = ModelObserver(pipeline=pipeline)
        obs.check_noise_invariance()

    def test_check_slice_invariance(self):
        """
        This is a complex test. Slice invariance will be triggered with a linear
        model, when 90% of the data is linearly separable, but 10% is displaced
        compared to the fit.

        We just do this for classification for ease, the observer runs
        irrespective of mode.
        """
        # Classification dataset
        x = np.linspace(0, 10, 100)
        y = np.concatenate(
            (
                np.zeros(48),
                np.ones(48),
                np.zeros(4),
            )
        )

        # Make pipeline and fit
        pipeline = Pipeline(n_grid_searches=0, document_results=False)
        pipeline._read_data(x, y)
        pipeline._mode_detector()
        pipeline._data_processing()
        pipeline._feature_processing()
        pipeline.conclude_fitting(
            model="LogisticRegression", params={}, feature_set="rf_increment"
        )

        # Observe
        obs = ModelObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning):
            obs.check_slice_invariance()

        # Should not trigger normally
        pipeline.best_model = RandomPredictor(mode="classification")
        obs = ModelObserver(pipeline=pipeline)
        obs.check_slice_invariance()

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_check_boosting_overfit(self, mode):
        if mode == "classification":
            x, y = make_classification(class_sep=0.5, n_samples=100)
        else:
            x, y = make_regression(noise=0.6, n_samples=100)

        # Make pipeline and simulate fit
        pipeline = Pipeline(n_grid_searches=0, cv_splits=2)
        pipeline._read_data(x, y)
        pipeline._mode_detector()
        n_estimators = 1000
        boost_kwargs = dict(
            n_estimators=n_estimators,
            l2_leaf_reg=0,
            early_stopping_rounds=n_estimators,
            use_best_model=False,
        )
        if mode == "classification":
            pipeline.best_model = CatBoostClassifier(**boost_kwargs)
        else:
            pipeline.best_model = CatBoostRegressor(**boost_kwargs)

        # Observer
        obs = ModelObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning):
            obs.check_boosting_overfit()



#  Copyright (c) 2022 by Amplo.

import pytest

from tests import make_data as _make_data
from tests import make_x_y as _make_x_y
from tests import rmfile, rmtree


@pytest.fixture(autouse=True)
def rmtree_automl():
    folder = "Auto_ML"
    rmtree(folder, must_exist=False)
    yield folder
    rmtree(folder, must_exist=False)


@pytest.fixture(autouse=True)
def rmfile_automl():
    file = "AutoML.log"
    yield file
    try:
        rmfile(file, must_exist=False)
    except PermissionError:
        pass


@pytest.fixture
def make_x_y(mode):
    yield _make_x_y(mode=mode)


@pytest.fixture
def make_data(mode, make_x_y, target="target"):
    yield _make_data(mode=mode, target=target)



#  Copyright (c) 2022 by Amplo.

import copy
import itertools
import json
import os
import time
import warnings
from datetime import datetime
from pathlib import Path
from typing import Tuple, Union

import joblib
import numpy as np
import pandas as pd
from shap import TreeExplainer
from sklearn import metrics
from sklearn.model_selection import KFold, StratifiedKFold
from tqdm import tqdm

import amplo
from amplo import utils
from amplo.automl.data_exploring import DataExplorer
from amplo.automl.data_processing import DataProcessor
from amplo.automl.data_sampling import DataSampler
from amplo.automl.drift_detection import DriftDetector
from amplo.automl.feature_processing import FeatureProcessor
from amplo.automl.interval_analysis import IntervalAnalyser
from amplo.automl.modelling import Modeller
from amplo.automl.sequencing import Sequencer
from amplo.classification.stacking import StackingClassifier
from amplo.grid_search import ExhaustiveGridSearch, HalvingGridSearch, OptunaGridSearch
from amplo.observation import DataObserver, ModelObserver
from amplo.regression.stacking import StackingRegressor
from amplo.validation import ModelValidator


class Pipeline:
    def __init__(self, **kwargs):
        """
        Automated Machine Learning Pipeline for tabular data.
        Designed for predictive maintenance applications, failure identification,
        failure prediction, condition monitoring, etc.

        Parameters
        ----------
        Main Parameters:
        main_dir : str, default: AutoML
            Main directory of Pipeline (for documentation)
        target : str
            Column name of the output/dependent/regressand variable.
        name : str, default: ''
            Name of the project (for documentation)
        version : str, default: detected
            Pipeline version (set automatically)
        mode : str, default: detected
            'classification' or 'regression'
        objective : str, default: neg_log_loss or mean_square_error
            from sklearn metrics and scoring

        Data Processor:
        int_cols : list of str, default: detected
            Column names of integer columns
        float_cols : list of str, default: detected
            Column names of float columns
        date_cols : list of str, default: detected
            Column names of datetime columns
        cat_cols : list of str, default: detected
            Column names of categorical columns
        missing_values : str, default: zero
            [DataProcessing] - 'remove', 'interpolate', 'mean' or 'zero'
        outlier_removal : str, default: clip
            [DataProcessing] - 'clip', 'boxplot', 'z-score' or 'none'
        z_score_threshold : int, default: 4
            [DataProcessing] If outlier_removal = 'z-score', the threshold is adaptable
        include_output : bool, default: False
            Whether to include output in the training data (sensible only with
            sequencing)

        Feature Processor:
        extract_features : bool, default: True
            Whether to use FeatureProcessing module
        information_threshold : float, default: 0.999
            [FeatureProcessing] Threshold for removing co-linear features
        feature_timeout : int, default: 3600
            [FeatureProcessing] Time budget for feature processing
        max_lags : int, default: 0
            [FeatureProcessing] Maximum lags for lagged features to analyse
        max_diff : int, default: 0
            [FeatureProcessing] Maximum differencing order for differencing features

        Interval Analyser:
        interval_analyse : bool, default: True
            Whether to use IntervalAnalyser module
            Note that this has no effect when data from ``self._read_data`` is not
            multi-indexed

        Sequencing:
        sequence : bool, default: False
            [Sequencing] Whether to use Sequence module
        seq_back : int or list of int, default: 1
            Input time indices
            If list -> includes all integers within the list
            If int -> includes that many samples back
        seq_forward : int or list of int, default: 1
            Output time indices
            If list -> includes all integers within the list.
            If int -> includes that many samples forward.
        seq_shift : int, default: 0
            Shift input / output samples in time
        seq_diff : str, default: 'none'
            Difference the input & output, 'none', 'diff' or 'log_diff'
        seq_flat : bool, default: True
            Whether to return a matrix (True) or Tensor (Flat)

        Modelling:
        standardize : bool, default: False
            Whether to standardize input/output data
        shuffle : bool, default: True
            Whether to shuffle the samples during cross-validation
        cv_splits : int, default: 10
            How many cross-validation splits to make
        store_models : bool, default: false
            Whether to store all trained model files

        Grid Search:
        grid_search_type : str, default: 'optuna'
            Grid search type to use. One of {'optuna', 'halving', 'exhaustive', None}.
        grid_search_time_budget : int, default: 3600
            Time budget for grid search (in seconds).
        n_grid_searches : int, default: 3
            Run grid search for the best `n_grid_searches` (model, feature set) pairs
            from initial modelling.
        n_trials_per_grid_search : int, default: 250
            Maximal number of trials/candidates for each grid search.

        Stacking:
        stacking : bool, default: False
            Whether to create a stacking model at the end

        Production:
        preprocess_function : str, default: None
            Add custom code for the prediction function, useful for production. Will
            be executed with exec, can be multiline. Uses data as input.

        Flags:
        logging_level  : int or str, optional
            Logging level for warnings, info, etc.
        plot_eda : bool, default: False
            Whether to run Exploratory Data Analysis
        process_data : bool, default: True
            Whether to force data processing
        no_dirs : bool, default: False
            Whether to create files or not
        verbose : int, default: 1
            Level of verbosity
        """

        # Set logger
        self.logger = utils.logging.logger

        # Copy arguments
        ##################
        # Main Settings
        self.main_dir = kwargs.get("main_dir", "Auto_ML/")
        self.target = kwargs.get("target", "")
        self.name = kwargs.get("name", "AutoML")
        self.version = kwargs.get("version", None)
        self.mode = kwargs.get("mode", None)
        self.objective = kwargs.get("objective", None)

        # Data Processor
        self.int_cols = kwargs.get("int_cols", None)
        self.float_cols = kwargs.get("float_cols", None)
        self.date_cols = kwargs.get("date_cols", None)
        self.cat_cols = kwargs.get("cat_cols", None)
        self.missing_values = kwargs.get("missing_values", "zero")
        self.outlier_removal = kwargs.get("outlier_removal", "clip")
        self.z_score_threshold = kwargs.get("z_score_threshold", 4)
        self.include_output = kwargs.get("include_output", False)

        # Balancer
        self.balance = kwargs.get("balance", True)

        # Feature Processor
        self.extract_features = kwargs.get("extract_features", True)
        self.information_threshold = kwargs.get("information_threshold", 0.999)
        self.feature_timeout = kwargs.get("feature_timeout", 3600)
        self.maxLags = kwargs.get("max_lags", 0)
        self.maxDiff = kwargs.get("max_diff", 0)

        # Interval Analyser
        self.use_interval_analyser = kwargs.get("interval_analyse", True)

        # Sequencer
        self.sequence = kwargs.get("sequence", False)
        self.sequence_back = kwargs.get("seq_back", 1)
        self.sequence_forward = kwargs.get("seq_forward", 1)
        self.sequence_shift = kwargs.get("seq_shift", 0)
        self.sequence_diff = kwargs.get("seq_diff", "none")
        self.sequence_flat = kwargs.get("seq_flat", True)

        # Modelling
        self.standardize = kwargs.get("standardize", False)
        self.cv_shuffle = kwargs.get("cv_shuffle", True)
        self.cv_splits = kwargs.get("cv_splits", 10)
        self.store_models = kwargs.get("store_models", False)

        # Grid Search Parameters
        self.grid_search_type = kwargs.get("grid_search_type", "optuna")
        self.grid_search_timeout = kwargs.get("grid_search_time_budget", 3600)
        self.n_grid_searches = kwargs.get("n_grid_searches", 3)
        self.n_trials_per_grid_search = kwargs.get("n_trials_per_grid_search", 250)

        # Stacking
        self.stacking = kwargs.get("stacking", False)

        # Production
        self.preprocess_function = kwargs.get("preprocess_function", None)

        # Flags
        self.plot_eda = kwargs.get("plot_eda", False)
        self.process_data = kwargs.get("process_data", True)
        self.verbose = kwargs.get("verbose", 1)
        self.no_dirs = kwargs.get("no_dirs", False)

        # Checks
        if self.mode not in [None, "regression", "classification"]:
            raise ValueError("Supported modes: regression, classification.")
        if not (0 < self.information_threshold < 1):
            raise ValueError("Information threshold needs to be within [0, 1].")
        if self.maxLags > 50:
            raise ValueError("Max_lags too big. Max 50.")
        if self.maxDiff > 5:
            raise ValueError("Max diff too big. Max 5.")
        self.grid_search_types = ["exhaustive", "halving", "optuna"]
        if (
            self.grid_search_type is not None
            and self.grid_search_type.lower() not in self.grid_search_types
        ):
            raise ValueError(f"Grid Search Type must be in {self.grid_search_types}")

        # Advices
        if self.include_output and not self.sequence:
            warnings.warn("Strongly advices to not include output without sequencing.")

        # Create dirs
        if not self.no_dirs:
            self._create_dirs()
            self._load_version()
        else:
            self.version = 1

        # Store Pipeline Settings
        self.settings = {"pipeline": kwargs}

        # Objective & Scorer
        if self.objective is not None:
            if not isinstance(self.objective, str):
                raise ValueError("Objective needs to be a string.")
            if self.objective not in metrics.SCORERS:
                raise NotImplementedError(
                    "Metric not supported, look at sklearn.metrics."
                )
            self.scorer = metrics.SCORERS[self.objective]
        else:
            self.scorer = None

        # Required sub-classes
        self.data_sampler = DataSampler()
        self.data_processor = DataProcessor()
        self.data_sequencer = Sequencer()
        self.feature_processor = FeatureProcessor()
        self.interval_analyser = IntervalAnalyser()
        self.drift_detector = DriftDetector()

        # Instance initiating
        self.best_model = None
        self.best_model_str = None
        self.best_feature_set = None
        self.best_score = None
        self._data = None
        self.feature_sets = None
        self.results = None
        self.n_classes = None
        self.is_fitted = False

        # Monitoring
        self._prediction_time = None
        self._main_predictors = None

    # User Pointing Functions
    def get_settings(self, version: int = None) -> dict:
        """
        Get settings to recreate fitted object.

        Parameters
        ----------
        version : int, optional
            Production version, defaults to current version
        """
        if version is None or version == self.version:
            assert self.is_fitted, "Pipeline not yet fitted."
            return self.settings
        else:
            settings_path = self.main_dir + f"Production/v{self.version}/Settings.json"
            assert Path(
                settings_path
            ).exists(), "Cannot load settings from nonexistent version"
            return json.load(open(settings_path, "r"))

    def load_settings(self, settings: dict):
        """
        Restores a pipeline from settings.

        Parameters
        ----------
        settings [dict]: Pipeline settings
        """
        # Set parameters
        settings["pipeline"]["no_dirs"] = True
        self.__init__(**settings["pipeline"])
        self.settings = settings
        self.data_processor.load_settings(settings["data_processing"])
        self.feature_processor.load_settings(settings["feature_processing"])
        # TODO: load_settings for IntervalAnalyser (not yet implemented)
        if "drift_detector" in settings:
            self.drift_detector = DriftDetector(
                num_cols=self.data_processor.float_cols + self.data_processor.int_cols,
                cat_cols=self.data_processor.cat_cols,
                date_cols=self.data_processor.date_cols,
            ).load_weights(settings["drift_detector"])

    def load_model(self, model: object):
        """
        Restores a trained model
        """
        assert type(model).__name__ == self.settings["model"]
        self.best_model = model
        self.is_fitted = True

    def fit(self, *args, **kwargs):
        """
        Fit the full AutoML pipeline.
            1. Prepare data for training
            2. Train / optimize models
            3. Prepare Production Files
                Nicely organises all required scripts / files to make a prediction

        Parameters
        ----------
        args
            For data reading - Propagated to `self.data_preparation`
        kwargs
            For data reading (propagated to `self.data_preparation`) AND
            for production filing (propagated to `self.conclude_fitting`)
        """
        # Starting
        self.logger.info(f"\n\n*** Starting Amplo AutoML - {self.name} ***\n\n")

        # Prepare data for training
        self.data_preparation(*args, **kwargs)

        # Train / optimize models
        self.model_training(**kwargs)

        # Conclude fitting
        self.conclude_fitting(**kwargs)

    def data_preparation(self, *args, **kwargs):
        """
        Prepare data for modelling
            1. Data Processing
                Cleans all the data. See @DataProcessing
            2. (optional) Exploratory Data Analysis
                Creates a ton of plots which are helpful to improve predictions manually
            3. Feature Processing
                Extracts & Selects. See @FeatureProcessing

        Parameters
        ----------
        args
            For data reading - Propagated to `self._read_data`
        kwargs
            For data reading - Propagated to `self._read_data`
        """
        # Reading data
        self._read_data(*args, **kwargs)

        # Detect mode (classification / regression)
        self._mode_detector()

        # Preprocess Data
        self._data_processing()

        # Check data
        obs = DataObserver(pipeline=self)
        obs.observe()

        # Fit Drift Detector to input
        num_cols = list(
            set(self.x).intersection(
                self.data_processor.float_cols + self.data_processor.int_cols
            )
        )
        date_cols = list(set(self.x).intersection(self.data_processor.date_cols))
        cat_cols = list(set(self.x) - set(num_cols + date_cols))
        self.drift_detector = DriftDetector(
            num_cols=num_cols, cat_cols=cat_cols, date_cols=date_cols
        )
        self.drift_detector.fit(self.x)

        # Run Exploratory Data Analysis
        self._eda()

        # Balance data
        self._data_sampling()

        # Sequence
        self._sequencing()

        # Interval-analyze data
        self._interval_analysis()

        # Extract and select features
        self._feature_processing()

        # Standardize
        # Standardizing assures equal scales, equal gradients and no clipping.
        # Therefore, it needs to be after sequencing & feature processing, as this
        # alters scales
        self._standardizing()

    def model_training(self, **kwargs):
        """Train models

        1. Initial Modelling
            Runs various models with default parameters for all feature sets
            If Sequencing is enabled, this is where it happens, as here, the feature
            set is generated.
        2. Grid Search
            Optimizes the hyperparameters of the best performing models
        3. (optional) Create Stacking model
        4. (optional) Create documentation

        Parameters
        ----------
        kwargs : optional
            Keyword arguments that will be passed to `self.grid_search`.
        """
        # Run initial models
        self._initial_modelling()

        # Optimize Hyper parameters
        self.grid_search(**kwargs)

        # Create stacking model
        self._create_stacking()

    def conclude_fitting(self, *, model=None, feature_set=None, params=None):
        """
        Prepare production files that are necessary to deploy a specific
        model / feature set combination

        Creates or modifies the following files
            - ``Model.joblib`` (production model)
            - ``Settings.json`` (model settings)

        Parameters
        ----------
        model : str or list of str, optional
            Model file for which to prepare production files.
        feature_set : str or list of str, optional
            Feature set for which to prepare production files.
        params : dict, optional
            Model parameters for which to prepare production files.
            Default: takes the best parameters
        """
        if not self.no_dirs:

            # Set up production path
            prod_dir = self.main_dir + f"Production/v{self.version}/"
            Path(prod_dir).mkdir(exist_ok=True)

            # Parse arguments
            self._parse_production_args(model, feature_set, params)

            # Verbose printing
            if self.verbose > 0:
                self.logger.info(
                    f"Preparing Production files for {self.best_model}, "
                    f"{self.best_feature_set}"
                )

            # Set best model (`self.bestModel`)
            self._prepare_production_model(prod_dir + "Model.joblib")

            # Set and store production settings
            self._prepare_production_settings(prod_dir + "Settings.json")

            # Observe production
            obs = ModelObserver(pipeline=self)
            obs.observe()
            self.settings["production_observation"] = obs.observations

        # Finish
        self.is_fitted = True
        self.logger.info("All done :)")
        utils.logging.remove_file_handler()

    def convert_data(
        self, x: pd.DataFrame, preprocess: bool = True
    ) -> Tuple[pd.DataFrame, pd.Series]:
        """
        Function that uses the same process as the pipeline to clean data.
        Useful if pipeline is pickled for production

        Parameters
        ----------
        data [pd.DataFrame]: Input features
        """
        # Convert to Pandas
        if isinstance(x, np.ndarray):
            x = pd.DataFrame(x, columns=[f"Feature_{i}" for i in range(x.shape[1])])

        # Custom code
        if self.preprocess_function is not None and preprocess:
            ex_globals = {"data": x}
            exec(self.preprocess_function, ex_globals)
            x = ex_globals["data"]

        # Process data
        x = self.data_processor.transform(x)

        # Drift Check
        self.drift_detector.check(x)

        # Split output
        y = None
        if self.target in x.keys():
            y = x[self.target]
            if not self.include_output:
                x = x.drop(self.target, axis=1)

        # Sequence
        if self.sequence:
            x, y = self.data_sequencer.convert(x, y)

        # Convert Features
        x = self.feature_processor.transform(
            x, feature_set=self.settings["feature_set"]
        )

        # Standardize
        if self.standardize:
            x, y = self._transform_standardize(x, y)

        # NaN test -- datetime should be taken care of by now
        if (
            x.astype("float32").replace([np.inf, -np.inf], np.nan).isna().sum().sum()
            != 0
        ):
            raise ValueError(
                f"Column(s) with NaN: {list(x.keys()[x.isna().sum() > 0])}"
            )

        # Return
        return x, y

    def predict(self, data: pd.DataFrame) -> np.ndarray:
        """
        Full script to make predictions. Uses 'Production' folder with defined or
        latest version.

        Parameters
        ----------
        data [pd.DataFrame]: data to do prediction on
        """
        start_time = time.time()
        assert self.is_fitted, "Pipeline not yet fitted."

        # Print
        if self.verbose > 0:
            self.logger.info(
                f"Predicting with {type(self.best_model).__name__}, v{self.version}"
            )

        # Convert
        x, y = self.convert_data(data)

        # Predict
        predictions = self.best_model.predict(x)
        if self.mode == "regression" and self.standardize:
            predictions = self._inverse_standardize(predictions)
        elif self.mode == "classification":
            predictions[:] = self.data_processor.decode_labels(
                predictions.astype(int), except_not_fitted=False
            )

        # Stop timer
        self._prediction_time = (time.time() - start_time) / len(x) * 1000

        # Calculate main predictors
        self._get_main_predictors(x)

        return predictions

    def predict_proba(self, data: pd.DataFrame) -> np.ndarray:
        """
        Returns probabilistic prediction, only for classification.

        Parameters
        ----------
        data [pd.DataFrame]: data to do prediction on
        """
        start_time = time.time()
        assert self.is_fitted, "Pipeline not yet fitted."
        assert (
            self.mode == "classification"
        ), "Predict_proba only available for classification"
        assert hasattr(
            self.best_model, "predict_proba"
        ), f"{type(self.best_model).__name__} has no attribute predict_proba"

        # Print
        if self.verbose > 0:
            self.logger.info(
                f"Predicting with {type(self.best_model).__name__}, v{self.version}"
            )

        # Convert data
        x, y = self.convert_data(data)

        # Predict
        prediction = self.best_model.predict_proba(x)

        # Stop timer
        self._prediction_time = (time.time() - start_time) / len(x) * 1000

        # Calculate main predictors
        self._get_main_predictors(x)

        return prediction

    # Fit functions
    def _read_data(self, x=None, y=None, *, data=None, **kwargs):
        """
        Reads and loads data into desired format.

        Expects to receive:
            1. Both, ``x`` and ``y`` (-> features and target), or
            2. Either ``x`` or ``data`` (-> dataframe or path to folder)

        Parameters
        ----------
        x : np.ndarray or pd.Series or pd.DataFrame or str or Path, optional
            x-data (input) OR acts as ``data`` parameter when param ``y`` is empty
        y : np.ndarray or pd.Series, optional
            y-data (target)
        data : pd.DataFrame or str or Path, optional
            Contains both, x and y, OR provides a path to folder structure
        kwargs : dict
            Not used, but necessary as we don't split .fit() kwargs

        Returns
        -------
        Pipeline
        """
        assert x is not None or data is not None, "No data provided"
        assert (x is not None) ^ (
            data is not None
        ), "Setting both, `x` and `data`, is ambiguous"

        # Labels are provided separately
        if y is not None:
            # Check data
            x = x if x is not None else data
            assert x is not None, "Parameter ``x`` is not set"
            assert isinstance(
                x, (np.ndarray, pd.Series, pd.DataFrame)
            ), "Unsupported data type for parameter ``x``"
            assert isinstance(
                y, (np.ndarray, pd.Series)
            ), "Unsupported data type for parameter ``y``"

            # Set target manually if not defined
            if not self.target:
                self.target = "target"

            # Parse x-data
            if isinstance(x, np.ndarray):
                x = pd.DataFrame(x)
            elif isinstance(x, pd.Series):
                x = pd.DataFrame(x)
            # Parse y-data
            if isinstance(y, np.ndarray):
                y = pd.Series(y, index=x.index)
            y.name = self.target

            # Check data
            assert all(x.index == y.index), "``x`` and ``y`` indices do not match"
            if self.target in x.columns:
                if any(x[self.target] != y):
                    msg = (
                        f"Target column coexists in both, x and y data, but have "
                        f"unequal content. Rename the column ``{self.target}`` in x."
                    )
                    raise ValueError(msg)
                data = x
            else:
                # Concatenate x and y
                data = pd.concat([x, y], axis=1)

        # Set data parameter in case it is provided through parameter ``x``
        data = data if data is not None else x
        metadata = None

        # A path was provided to read out (multi-indexed) data
        if isinstance(data, (str, Path)):
            # Set target manually if not defined
            if not self.target:
                self.target = "target"
            # Parse data
            data, metadata = utils.io.merge_logs(data, self.target)

        # Test data
        assert self.target, "No target string provided"
        assert self.target in data.columns, "Target column missing"
        assert (
            len(data.columns) == data.columns.nunique()
        ), "Columns have no unique names"

        # Save data
        clean_target = utils.clean_feature_name(self.target)
        data.rename(columns={self.target: clean_target}, inplace=True)
        self.target = clean_target
        self._set_data(data)

        # Store metadata in settings
        self.settings["file_metadata"] = metadata or dict()

        return self

    def has_new_training_data(self):
        # Return True if no previous version exists
        if self.version == 1:
            return True

        # Get previous and current file metadata
        curr_metadata = self.settings["file_metadata"]
        last_metadata = self.get_settings(self.version - 1)["file_metadata"]

        # Check each settings file
        for file_id in curr_metadata:
            # Get file specific metadata
            curr = curr_metadata[file_id]
            last = last_metadata.get(file_id, dict())
            # Compare metadata
            same_folder = curr["folder"] == last.get("folder")
            same_file = curr["file"] == last.get("file")
            same_mtime = curr["last_modified"] == last.get("last_modified")
            if not all([same_folder, same_file, same_mtime]):
                return False

        return True

    def _mode_detector(self):
        """
        Detects the mode (Regression / Classification)
        """
        # Only run if mode is not provided
        if self.mode is None:

            # Classification if string
            if self.y.dtype == str or self.y.nunique() < 0.1 * len(self.data):
                self.mode = "classification"
                self.objective = self.objective or "neg_log_loss"

            # Else regression
            else:
                self.mode = "regression"
                self.objective = self.objective or "neg_mean_absolute_error"

            # Set scorer
            self.scorer = metrics.SCORERS[self.objective]

            # Copy to settings
            self.settings["pipeline"]["mode"] = self.mode
            self.settings["pipeline"]["objective"] = self.objective

            # Print
            if self.verbose > 0:
                self.logger.info(
                    f"Setting mode to {self.mode} & objective to {self.objective}."
                )
        return

    def _data_processing(self):
        """
        Organises the data cleaning. Heavy lifting is done in self.dataProcessor,
        but settings etc. needs to be organised.
        """
        self.data_processor = DataProcessor(
            target=self.target,
            int_cols=self.int_cols,
            float_cols=self.float_cols,
            date_cols=self.date_cols,
            cat_cols=self.cat_cols,
            include_output=True,
            missing_values=self.missing_values,
            outlier_removal=self.outlier_removal,
            z_score_threshold=self.z_score_threshold,
        )

        # Set paths
        extracted_data_path = self.main_dir + f"Data/Extracted_v{self.version}.csv"
        data_path = self.main_dir + f"Data/Cleaned_v{self.version}.csv"
        settings_path = self.main_dir + f"Settings/Cleaning_v{self.version}.json"

        # Only load settings if feature processor is done already.
        if Path(extracted_data_path).exists() and Path(settings_path).exists():
            self.settings["data_processing"] = json.load(open(settings_path, "r"))
            self.data_processor.load_settings(self.settings["data_processing"])

        # Else, if data processor is done, load settings & data
        elif Path(data_path).exists() and Path(settings_path).exists():
            # Load data
            data = self._read_csv(data_path)
            self._set_data(data)

            # Load settings
            self.settings["data_processing"] = json.load(open(settings_path, "r"))
            self.data_processor.load_settings(self.settings["data_processing"])

            if self.verbose > 0:
                self.logger.info("Loaded Cleaned Data")

        # Else, run the data processor
        else:
            # Cleaning
            data = self.data_processor.fit_transform(self.data)

            # Update pipeline
            self._set_data(data)

            # Store data
            self._write_csv(self.data, data_path)

            # Clean up Extracted previous version
            if os.path.exists(
                self.main_dir + f"Data/Extracted_v{self.version - 1}.csv"
            ):
                os.remove(self.main_dir + f"Data/Extracted_v{self.version - 1}.csv")

            # Save settings
            self.settings["data_processing"] = self.data_processor.get_settings()
            if not self.no_dirs:
                json.dump(self.settings["data_processing"], open(settings_path, "w"))

        # If no columns were provided, load them from data processor
        if self.date_cols is None:
            self.date_cols = self.settings["data_processing"]["date_cols"]
        if self.int_cols is None:
            self.int_cols = self.settings["data_processing"]["int_cols"]
        if self.float_cols is None:
            self.float_cols = self.settings["data_processing"]["float_cols"]
        if self.cat_cols is None:
            self.cat_cols = self.settings["data_processing"]["cat_cols"]

        # Assert classes in case of classification
        if self.mode == "classification":
            self.n_classes = self.y.nunique()
            if self.n_classes >= 50:
                warnings.warn(
                    "More than 20 classes, "
                    "you may want to reconsider classification mode"
                )
            if set(self.y) != set([i for i in range(len(set(self.y)))]):
                raise ValueError("Classes should be [0, 1, ...]")

    def _eda(self):
        if self.plot_eda:
            self.logger.info("Starting Exploratory Data Analysis")
            eda = DataExplorer(
                self.x,
                y=self.y,
                mode=self.mode,
                folder=self.main_dir,
                version=self.version,
            )
            eda.run()

    def _data_sampling(self):
        """
        Only run for classification problems. Balances the data using imblearn.
        Does not guarantee to return balanced classes. (Methods are data dependent)
        """
        self.data_sampler = DataSampler(
            method="both",
            margin=0.1,
            cv_splits=self.cv_splits,
            shuffle=self.cv_shuffle,
            fast_run=False,
            objective=self.objective,
        )

        # Set paths
        data_path = self.main_dir + f"Data/Balanced_v{self.version}.csv"

        # Only necessary for classification
        if self.mode == "classification" and self.balance:

            if Path(data_path).exists():
                # Load data
                data = self._read_csv(data_path)
                self._set_data(data)

                if self.verbose > 0:
                    self.logger.info("Loaded Balanced data")

            else:
                # Fit and resample
                self.logger.info("Resampling data")
                x, y = self.data_sampler.fit_resample(self.x, self.y)

                # Store
                self._set_xy(x, y)
                self._write_csv(self.data, data_path)

    def _sequencing(self):
        """
        Sequences the data. Useful mostly for problems where older samples play a role
        in future values. The settings of this module are NOT AUTOMATIC
        """
        self.data_sequencer = Sequencer(
            back=self.sequence_back,
            forward=self.sequence_forward,
            shift=self.sequence_shift,
            diff=self.sequence_diff,
        )

        # Set paths
        data_path = self.main_dir + f"Data/Sequence_v{self.version}.csv"

        if self.sequence:

            if Path(data_path).exists():
                # Load data
                data = self._read_csv(data_path)
                self._set_data(data)

                if self.verbose > 0:
                    self.logger.info("Loaded Extracted Features")

            else:
                # Sequencing
                self.logger.info("Sequencing data")
                x, y = self.data_sequencer.convert(self.x, self.y)

                # Store
                self._set_xy(x, y)
                self._write_csv(self.data, data_path)

    def _feature_processing(self):
        """
        Organises feature processing. Heavy lifting is done in self.featureProcessor,
        but settings, etc. needs to be organised.
        """
        self.feature_processor = FeatureProcessor(
            mode=self.mode,
            is_temporal=None,
            extract_features=self.extract_features,
            collinear_threshold=self.information_threshold,
        )

        # Set paths
        data_path = self.main_dir + f"Data/Extracted_v{self.version}.csv"
        settings_path = self.main_dir + f"Settings/Extracting_v{self.version}.json"

        if Path(data_path).exists() and Path(settings_path).exists():
            # Loading data
            data = self._read_csv(data_path)
            self._set_data(data)

            # Loading settings
            self.settings["feature_processing"] = json.load(open(settings_path, "r"))
            self.feature_processor.load_settings(self.settings["feature_processing"])
            self.feature_sets = self.settings["feature_processing"]["feature_sets_"]

            if self.verbose > 0:
                self.logger.info("Loaded Extracted Features")

        else:
            self.logger.info("Starting Feature Processor")

            # Transform data.  Note that y also needs to be transformed in the
            # case when we're using the temporal feature processor (pooling).
            x = self.feature_processor.fit_transform(self.x, self.y)
            y = self.feature_processor.transform_target(self.y)
            self.feature_sets = self.feature_processor.feature_sets_

            # Store data
            self._set_xy(x, y)
            self._write_csv(self.data, data_path)

            # Cleanup Cleaned_vx.csv
            if os.path.exists(self.main_dir + f"Data/Cleaned_v{self.version}.csv"):
                os.remove(self.main_dir + f"Data/Cleaned_v{self.version}.csv")

            # Save settings
            self.settings["feature_processing"] = self.feature_processor.get_settings()
            if not self.no_dirs:
                json.dump(self.settings["feature_processing"], open(settings_path, "w"))

    def _interval_analysis(self):
        """
        Interval-analyzes the data using ``amplo.auto_ml.interval_analysis``
        or resorts to pre-computed data, if present.
        """
        # Skip analysis when analysis is not possible and/or not desired
        is_multi_index = len(self.x.index.names) == 2
        if not all(
            [self.use_interval_analyser, is_multi_index, self.mode == "classification"]
        ):
            return

        self.interval_analyser = IntervalAnalyser(target=self.target)

        # Set paths
        data_path = self.main_dir + f"Data/Interval_Analyzed_v{self.version}.csv"
        # settings_path = self.main_dir +
        # f'Settings/Interval_Analysis_v{self.version}.json'

        if Path(data_path).exists():  # TODO: and Path(settings_path).exists():
            # Load data
            data = self._read_csv(data_path)
            self._set_data(data)

            # TODO implement `IntervalAnalyser.load_settings` and add to
            #  `self.load_settings`
            # # Load settings
            # self.settings['interval_analysis'] = json.load(open(settings_path, 'r'))
            # self.intervalAnalyser.load_settings(self.settings['interval_analysis'])

            if self.verbose > 0:
                self.logger.info("Loaded interval-analyzed data")

        else:
            self.logger.info("Interval-analyzing data")

            # Transform data
            data = self.interval_analyser.fit_transform(self.x, self.y)

            # Store data
            self._set_data(data)
            self._write_csv(self.data, data_path)

            # TODO implement `IntervalAnalyser.get_settings` and add to
            #  `self.get_settings`
            # # Save settings
            # self.settings['interval_analysis'] = self.intervalAnalyser.get_settings()
            # json.dump(self.settings['interval_analysis'], open(settings_path, 'w'))

    def _standardizing(self):
        """
        Wrapper function to determine whether to fit or load
        """
        # Return if standardize is off
        if not self.standardize:
            return

        # Set paths
        settings_path = self.main_dir + f"Settings/Standardize_v{self.version}.json"

        if Path(settings_path).exists():
            # Load data
            self.settings["standardize"] = json.load(open(settings_path, "r"))

        else:
            # Fit data
            self._fit_standardize(self.x, self.y)

            # Store Settings
            json.dump(self.settings["standardize"], open(settings_path, "w"))

        # Transform data
        x, y = self._transform_standardize(self.x, self.y)
        self._set_xy(x, y)

    def _initial_modelling(self):
        """
        Runs various models to see which work well.
        """

        # Set paths
        results_path = Path(self.main_dir) / "Results.csv"

        # Load existing results
        if results_path.exists():

            # Load results
            self.results = pd.read_csv(results_path)

            # Printing here as we load it
            results = self.results[
                np.logical_and(
                    self.results["version"] == self.version,
                    self.results["type"] == "Initial modelling",
                )
            ]
            for fs in set(results["dataset"]):
                self.logger.info(
                    f"Initial Modelling for {fs} ({len(self.feature_sets[fs])})"
                )
                fsr = results[results["dataset"] == fs]
                for i in range(len(fsr)):
                    row = fsr.iloc[i]
                    self.logger.info(
                        f'{row["model"].ljust(40)} {self.objective}: '
                        f'{row["mean_objective"]:.4f} \u00B1 {row["std_objective"]:.4f}'
                    )

        # Check if this version has been modelled
        if self.results is None or self.version not in self.results["version"].values:

            # Iterate through feature sets
            for feature_set, cols in self.feature_sets.items():

                # Skip empty sets
                if len(cols) == 0:
                    self.logger.info(f"Skipping {feature_set} features, empty set")
                    continue
                self.logger.info(
                    f"Initial Modelling for {feature_set} features ({len(cols)})"
                )

                # Do the modelling
                modeller = Modeller(
                    mode=self.mode,
                    store_models=self.store_models,
                    cv=self.cv,
                    objective=self.objective,
                    dataset=feature_set,
                    store_results=False,
                    folder=self.main_dir + "Models/",
                )
                results = modeller.fit(self.x[cols], self.y)

                # Add results to memory
                results["type"] = "Initial modelling"
                results["version"] = self.version
                if self.results is None:
                    self.results = results
                else:
                    self.results = pd.concat([self.results, results])

            # Save results
            self._write_csv(self.results, results_path)

    def grid_search(self, model=None, feature_set=None, parameter_set=None):
        """
        Runs a grid search.

        By default, takes ``self.results`` and runs for the top ``n =
        self.n_grid_searches`` optimizations. There is the option to provide ``model``
        and ``feature_set``, but **both** have to be provided. In this case, the model
        and dataset combination will be optimized.

        Implemented types, Exhaustive, Halving, Optuna.

        Parameters
        ----------
        model : list of (str or object) or object or str, optional
            Which model to run grid search for.
        feature_set : list of str or str, optional
            Which feature set to run gid search for. Must be provided when `model` is
            not None.
            Options: {rf_threshold, rf_increment, shap_threshold, shap_increment}
        parameter_set : dict, optional
            Parameter grid to optimize over.

        Notes
        -----
        When both parameters, ``model`` and ``feature_set``, are provided, the grid
        search behaves as follows:
            - When both parameters are either of dtype ``str`` or have the same length,
              then grid search will treat them as pairs.
            - When one parameter is an iterable and the other parameter is either a
              string or an iterable of different length, then grid search will happen
              for each unique combination of these parameters.
        """

        # Skip grid search and set best initial model as best grid search parameters
        if self.grid_search_type is None or self.n_grid_searches == 0:
            best_initial_model = self._sort_results(
                self.results[self.results["version"] == self.version]
            ).iloc[:1]
            best_initial_model["type"] = "Hyper Parameter"
            self.results = pd.concat(
                [self.results, best_initial_model], ignore_index=True
            )
            return self

        # Define models
        if model is None:
            # Run through first best initial models (n = self.n_grid_searches)
            selected_results = self.sort_results(
                self.results[
                    np.logical_and(
                        self.results["type"] == "Initial modelling",
                        self.results["version"] == self.version,
                    )
                ]
            ).iloc[: self.n_grid_searches]
            models = [
                utils.get_model(model_name, mode=self.mode, samples=len(self.x))
                for model_name in selected_results["model"]
            ]
            feature_sets = selected_results["dataset"]

        elif feature_set is None:
            raise AttributeError(
                "When `model` is provided, `feature_set` cannot be None. "
                "Provide either both params or neither of them."
            )

        else:
            models = (
                [utils.get_model(model, mode=self.mode, samples=len(self.x))]
                if isinstance(model, str)
                else [model]
            )
            feature_sets = (
                [feature_set] if isinstance(feature_set, str) else list(feature_set)
            )
            if len(models) != len(feature_sets):
                # Create each combination
                combinations = list(
                    itertools.product(np.unique(models), np.unique(feature_sets))
                )
                models = [elem[0] for elem in combinations]
                feature_sets = [elem[1] for elem in combinations]

        # Iterate and grid search over each pair of model and feature_set
        for model, feature_set in zip(models, feature_sets):

            # Organise existing model results
            m_results = self.results[
                np.logical_and(
                    self.results["model"] == type(model).__name__,
                    self.results["version"] == self.version,
                )
            ]
            m_results = self._sort_results(
                m_results[m_results["dataset"] == feature_set]
            )

            # Skip grid search if optimized model already exists
            if not ("Hyper Parameter" == m_results["type"]).any():
                # Run grid search for model
                grid_search_results = self._grid_search_iteration(
                    model, parameter_set, feature_set
                )
                grid_search_results = self.sort_results(grid_search_results)

                # Store results
                grid_search_results["version"] = self.version
                grid_search_results["dataset"] = feature_set
                grid_search_results["type"] = "Hyper Parameter"
                self.results = pd.concat(
                    [self.results, grid_search_results], ignore_index=True
                )
                self.results.to_csv(self.main_dir + "Results.csv", index=False)

        return self

    def _create_stacking(self):
        """
        Based on the best performing models, in addition to cheap models based on very
        different assumptions, a stacking model is optimized to enhance/combine the
        performance of the models.
        --> should contain a large variety of models
        --> classifiers need predict_proba
        --> level 1 needs to be ordinary least squares
        """
        if self.stacking:
            self.logger.info("Creating Stacking Ensemble")

            # Select feature set that has been picked most often for hyper parameter
            # optimization
            results = self._sort_results(
                self.results[
                    np.logical_and(
                        self.results["type"] == "Hyper Parameter",
                        self.results["version"] == self.version,
                    )
                ]
            )
            feature_set = results["dataset"].value_counts().index[0]
            results = results[results["dataset"] == feature_set]
            self.logger.info("Selected Stacking feature set: {}".format(feature_set))

            # Create Stacking Model Params
            n_stacking_models = 3
            stacking_models_str = results["model"].unique()[:n_stacking_models]
            stacking_models_params = [
                utils.io.parse_json(
                    results.iloc[np.where(results["model"] == sms)[0][0]]["params"]
                )
                for sms in stacking_models_str
            ]
            stacking_models = dict(
                [
                    (sms, stacking_models_params[i])
                    for i, sms in enumerate(stacking_models_str)
                ]
            )
            self.logger.info("Stacked models: {}".format(list(stacking_models.keys())))

            # Add samples & Features
            stacking_models["n_samples"], stacking_models["n_features"] = self.x.shape

            # Prepare Stack
            if self.mode == "regression":
                stack = StackingRegressor(**stacking_models)

            elif self.mode == "classification":
                stack = StackingClassifier(**stacking_models)
            else:
                raise NotImplementedError("Unknown mode")

            # Cross Validate
            x, y = self.x[self.feature_sets[feature_set]].to_numpy(), self.y.to_numpy()
            score = []
            times = []
            for (t, v) in tqdm(self.cv.split(x, y)):
                start_time = time.time()
                xt, xv, yt, yv = x[t], x[v], y[t].reshape((-1)), y[v].reshape((-1))
                model = copy.deepcopy(stack)
                model.fit(xt, yt)
                score.append(self.scorer(model, xv, yv))
                times.append(time.time() - start_time)

            # Output Results
            self.logger.info("Stacking result:")
            self.logger.info(
                f"{self.objective}:     {np.mean(score):.2f} \u00B1 {np.std(score):.2f}"
            )
            self.results = self.results.append(
                {
                    "date": datetime.today().strftime("%d %b %y"),
                    "model": type(stack).__name__,
                    "dataset": feature_set,
                    "params": json.dumps(stack.get_params()),
                    "mean_objective": np.mean(score),
                    "std_objective": np.std(score),
                    "mean_time": np.mean(times),
                    "std_time": np.std(times),
                    "version": self.version,
                    "type": "Stacking",
                },
                ignore_index=True,
            )
            self.results.to_csv(self.main_dir + "Results.csv", index=False)

    def _parse_production_args(self, model=None, feature_set=None, params=None):
        """
        Parse production arguments. Selects the best model, feature set and parameter
        combination.

        Parameters
        ----------
        model : str or list of str, optional
            Model constraint(s). In case list is provided, the pipeline needs to be
            fitted.
        feature_set : str or list of str, optional
            Feature set constraint(s). In case list is provided, the pipeline needs to
            be fitted.
        params : dict, optional
            Parameter constraint(s)
        Returns
        -------
        model : str
            Best model given the `model` restraint(s).
        feature_set : str
            Best feature set given the `feature_set` restraint(s).
        params : dict
            Best model parameters given the `params` restraint(s).
        """
        if self.results is None and (
            model is None or params is None or feature_set is None
        ):
            raise ValueError(
                "Pipeline not fitted and no model, params or feature set provided."
            )

        if model is not None and not isinstance(model, str):
            # TODO: This issue is linked with AML-103 (in Jira)
            #  1. Add to method docstring that it accepts a model instance, too
            #  2. Change `if`-case to a single `isinstance(model, BasePredictor)`
            # Get model name
            model = type(model).__name__

        # Get results of current version
        results = (
            self._sort_results(self.results[self.results["version"] == self.version])
            if self.results is not None
            else None
        )

        # Filter results for model
        if model is not None and results is not None:
            # Enforce list
            if isinstance(model, str):
                model = [model]

            # Filter results
            results = self._sort_results(results[results["model"].isin(model)])

        # filter results for feature set
        if feature_set is not None and results is not None:
            if isinstance(feature_set, str):
                feature_set = [feature_set]
            # Filter results
            results = self._sort_results(results[results["dataset"].isin(feature_set)])

        # Get best parameters
        if params is None and results is not None:
            params = results.iloc[0]["params"]
        elif params is None:
            params = {}

        # Find the best allowed arguments
        self.best_model_str = model if results is None else results.iloc[0]["model"]
        self.best_feature_set = (
            feature_set if results is None else results.iloc[0]["dataset"]
        )
        self.best_params = utils.io.parse_json(params)
        self.best_score = 0 if results is None else results.iloc[0]["worst_case"]

        return self

    def _prepare_production_model(self, model_path):
        """
        Prepare and store `self.bestModel` for production

        Parameters
        ----------
        model_path : str or Path
            Where to store model for production

        Returns
        -------
        """
        model_path = Path(model_path)

        # Make model
        if "Stacking" in self.best_model_str:
            # Create stacking
            if self.mode == "regression":
                self.best_model = StackingRegressor(
                    n_samples=len(self.x), n_features=len(self.x.keys())
                )
            elif self.mode == "classification":
                self.best_model = StackingClassifier(
                    n_samples=len(self.x), n_features=len(self.x.keys())
                )
            else:
                raise NotImplementedError("Mode not set")
        else:
            # Take model as is
            self.best_model = utils.get_model(
                self.best_model_str, mode=self.mode, samples=len(self.x)
            )

        # Set params, train
        self.best_model.set_params(**self.best_params)
        self.best_model.fit(self.x[self.feature_sets[self.best_feature_set]], self.y)

        # Save model
        joblib.dump(self.best_model, model_path)

        self.best_score = self.scorer(
            self.best_model,
            self.x[self.feature_sets[self.best_feature_set]],
            self.y,
        )
        if self.verbose > 0:
            self.logger.info(
                f"Model fully fitted, in-sample {self.objective}: {self.best_score:4f}"
            )

        return

    def _prepare_production_settings(self, settings_path):
        """
        Prepare `self.settings` for production and dump to file

        Parameters
        ----------
        settings_path : str or Path
            Where to save settings for production
        """
        assert self.best_model is not None, "`self.bestModel` is not yet prepared"
        settings_path = Path(settings_path)

        # Update pipeline settings
        self.settings["version"] = self.version
        self.settings["pipeline"]["verbose"] = self.verbose
        self.settings["model"] = self.best_model_str
        self.settings["params"] = self.best_params
        self.settings["feature_set"] = self.best_feature_set
        self.settings["features"] = self.feature_sets[self.best_feature_set]
        self.settings["best_score"] = self.best_score
        self.settings["amplo_version"] = (
            amplo.__version__ if hasattr(amplo, "__version__") else "dev"
        )

        # Validation
        validator = ModelValidator(
            cv_splits=self.cv_splits,
            cv_shuffle=self.cv_shuffle,
            verbose=self.verbose,
        )
        self.settings["validation"] = validator.validate(
            model=self.best_model, x=self.x, y=self.y, mode=self.mode
        )

        # Prune Data Processor
        required_features = self.feature_processor.get_required_columns(
            self.best_feature_set
        )
        self.data_processor.prune_features(required_features)
        self.settings["data_processing"] = self.data_processor.get_settings()

        # Fit Drift Detector to output and store settings
        self.drift_detector.fit_output(
            self.best_model, self.x[self.feature_sets[self.best_feature_set]]
        )
        self.settings["drift_detector"] = self.drift_detector.get_weights()

        # Save settings
        json.dump(self.settings, open(settings_path, "w"), indent=4)

    # Getter Functions / Properties
    @property
    def cv(self):
        """
        Gives the Cross Validation scheme

        Returns
        -------
        cv : sklearn.model_selection._search.BaseSearchCV
            The cross validator
        """
        # Regression
        if self.mode == "regression":
            return KFold(
                n_splits=self.cv_splits,
                shuffle=self.cv_shuffle,
                random_state=83847939 if self.cv_shuffle else None,
            )

        # Classification
        if self.mode == "classification":
            return StratifiedKFold(
                n_splits=self.cv_splits,
                shuffle=self.cv_shuffle,
                random_state=83847939 if self.cv_shuffle else None,
            )

    @property
    def data(self) -> Union[None, pd.DataFrame]:
        return self._data

    @property
    def x(self) -> pd.DataFrame:
        if self.data is None:
            raise AttributeError("Data is None")
        if self.include_output:
            return self.data
        return self.data.drop(self.target, axis=1)

    @property
    def y(self):
        if self.data is None:
            raise AssertionError("`self.data` is empty. Set a value with `set_data`")
        return self.data[self.target]

    @property
    def y_orig(self):
        enc_labels = self.y
        dec_labels = self.data_processor.decode_labels(
            enc_labels, except_not_fitted=False
        )
        return pd.Series(dec_labels, name=self.target, index=enc_labels.index)

    # Setter Functions
    def _set_data(self, new_data: pd.DataFrame):
        assert isinstance(new_data, pd.DataFrame), "Invalid data type"
        assert self.target in new_data, "No target column present"
        assert len(new_data.columns) > 1, "No feature column present"
        self._data = new_data

    def _set_xy(
        self,
        new_x: Union[np.ndarray, pd.DataFrame],
        new_y: Union[np.ndarray, pd.Series],
    ):
        if not isinstance(new_y, pd.Series):
            new_y = pd.Series(new_y, name=self.target)
        new_data = pd.concat([new_x, new_y], axis=1)
        self._set_data(new_data)

    def _set_x(self, new_x: Union[np.ndarray, pd.DataFrame]):
        # Convert to dataframe
        if isinstance(new_x, np.ndarray):
            old_x = self.data.drop(self.target, axis=1)
            old_x_shape = old_x.shape
            old_x_index = old_x.index
            old_x_columns = old_x.columns
            del old_x

            if new_x.shape == old_x_shape:
                new_x = pd.DataFrame(new_x, index=old_x_index, columns=old_x_columns)
            else:
                warnings.warn(
                    "Old x-data has more/less columns than new x-data. "
                    "Setting dummy feature names..."
                )
                columns = [f"Feature_{i}" for i in range(new_x.shape[1])]
                new_x = pd.DataFrame(new_x, index=old_x_index, columns=columns)

        elif not isinstance(new_x, pd.DataFrame):
            raise ValueError(f"Invalid dtype for new x data: {type(new_x)}")

        # Assert that target is not in x-data
        if self.target in new_x:
            raise AttributeError("Target column name should not be in x-data")

        # Set data
        self._data = pd.concat([new_x, self.y], axis=1)

    def set_y(self, new_y: Union[np.ndarray, pd.Series]):
        self._data[self.target] = new_y

    # Support Functions
    @staticmethod
    def _read_csv(data_path) -> pd.DataFrame:
        """
        Read data from given path and set index or multi-index

        Parameters
        ----------
        data_path : str or Path
        """
        assert Path(data_path).suffix == ".csv", "Expected a *.csv path"

        data = pd.read_csv(data_path)

        if {"index", "log"}.issubset(data.columns):
            # Multi-index: case when IntervalAnalyser was used
            index = ["log", "index"]
        elif "index" in data.columns:
            index = ["index"]
        else:
            raise IndexError(
                "No known index was found. "
                "Expected to find at least a column named `index`."
            )
        return data.set_index(index)

    def _write_csv(self, data, data_path):
        """
        Write data to given path and set index if needed.

        Parameters
        ----------
        data : pd.DataFrame or pd.Series
        data_path : str or Path
        """
        assert Path(data_path).suffix == ".csv", "Expected a *.csv path"

        # Set single-index if not already present
        if len(data.index.names) == 1 and data.index.name is None:
            data.index.name = "index"

        # Raise error if unnamed index is present
        if None in data.index.names:
            raise IndexError(
                f"Found an unnamed index column ({list(data.index.names)})."
            )

        # Write data
        if not self.no_dirs:
            data.to_csv(data_path)

    def _load_version(self):
        """
        Upon start, loads version
        """
        # No need if version is set
        if self.version is not None:
            return

        # Find production folders
        completed_versions = len(os.listdir(self.main_dir + "Production"))
        self.version = completed_versions + 1

        if self.verbose > 0:
            self.logger.info(f"Setting Version {self.version}")

    def _create_dirs(self):
        folders = [
            "",
            "EDA",
            "Data",
            "Features",
            "Documentation",
            "Production",
            "Settings",
        ]
        for folder in folders:
            if not os.path.exists(self.main_dir + folder):
                os.makedirs(self.main_dir + folder)

    def sort_results(self, results: pd.DataFrame) -> pd.DataFrame:
        return self._sort_results(results)

    def _fit_standardize(self, x: pd.DataFrame, y: pd.Series):
        """
        Fits a standardization parameters and returns the transformed data
        """

        # Fit Input
        cat_cols = [
            k
            for lst in self.settings["data_processing"]["dummies"].values()
            for k in lst
        ]
        features = [
            k for k in x.keys() if k not in self.date_cols and k not in cat_cols
        ]
        means_ = x[features].mean(axis=0)
        stds_ = x[features].std(axis=0)
        stds_[stds_ == 0] = 1
        settings = {
            "input": {
                "features": features,
                "means": means_.to_list(),
                "stds": stds_.to_list(),
            }
        }

        # Fit Output
        if self.mode == "regression":
            std = y.std()
            settings["output"] = {
                "mean": y.mean(),
                "std": std if std != 0 else 1,
            }

        self.settings["standardize"] = settings

    def _transform_standardize(
        self, x: pd.DataFrame, y: pd.Series
    ) -> Tuple[pd.DataFrame, pd.Series]:
        """
        Standardizes the input and output with values from settings.

        Parameters
        ----------
        x [pd.DataFrame]: Input data
        y [pd.Series]: Output data
        """
        # Input
        assert self.settings["standardize"], "Standardize settings not found."

        # Pull from settings
        features = self.settings["standardize"]["input"]["features"]
        means = self.settings["standardize"]["input"]["means"]
        stds = self.settings["standardize"]["input"]["stds"]

        # Filter if not all features are present
        if len(x.keys()) < len(features):
            indices = [
                [i for i, j in enumerate(features) if j == k][0] for k in x.keys()
            ]
            features = [features[i] for i in indices]
            means = [means[i] for i in indices]
            stds = [stds[i] for i in indices]

        # Transform Input
        x[features] = (x[features] - means) / stds

        # Transform output (only with regression)
        if self.mode == "regression":
            y = (y - self.settings["standardize"]["output"]["mean"]) / self.settings[
                "standardize"
            ]["output"]["std"]

        return x, y

    def _inverse_standardize(self, y: pd.Series) -> pd.Series:
        """
        For predictions, transform them back to application scales.
        Parameters
        ----------
        y [pd.Series]: Standardized output

        Returns
        -------
        y [pd.Series]: Actual output
        """
        assert self.settings["standardize"], "Standardize settings not found"
        return (
            y * self.settings["standardize"]["output"]["std"]
            + self.settings["standardize"]["output"]["mean"]
        )

    @staticmethod
    def _sort_results(results: pd.DataFrame) -> pd.DataFrame:
        return results.sort_values("worst_case", ascending=False)

    def _get_best_params(self, model, feature_set: str) -> dict:
        # Filter results for model and version
        results = self.results[
            np.logical_and(
                self.results["model"] == type(model).__name__,
                self.results["version"] == self.version,
            )
        ]

        # Filter results for feature set & sort them
        results = self._sort_results(results[results["dataset"] == feature_set])

        # Warning for unoptimized results
        if "Hyper Parameter" not in results["type"].values:
            warnings.warn("Hyper parameters not optimized for this combination")

        # Parse & return best parameters (regardless of if it's optimized)
        return utils.io.parse_json(results.iloc[0]["params"])

    def _grid_search_iteration(
        self, model, parameter_set: Union[str, None], feature_set: str
    ):
        """
        INTERNAL | Grid search for defined model, parameter set and feature set.
        """
        self.logger.info(
            f"\nStarting Hyper Parameter Optimization for {type(model).__name__} on "
            f"{feature_set} features ({len(self.x)} samples, "
            f"{len(self.feature_sets[feature_set])} features)"
        )

        # Select right hyper parameter optimizer
        if self.grid_search_type == "exhaustive":
            grid_search = ExhaustiveGridSearch(
                model,
                params=parameter_set,
                cv=self.cv,
                scoring=self.objective,
                n_trials=self.n_trials_per_grid_search,
                timeout=self.grid_search_timeout,
                verbose=self.verbose,
            )
        elif self.grid_search_type == "halving":
            grid_search = HalvingGridSearch(
                model,
                params=parameter_set,
                cv=self.cv,
                scoring=self.objective,
                n_trials=self.n_trials_per_grid_search,
                verbose=self.verbose,
            )
        elif self.grid_search_type == "optuna":
            grid_search = OptunaGridSearch(
                model,
                timeout=self.grid_search_timeout,
                cv=self.cv,
                n_trials=self.n_trials_per_grid_search,
                scoring=self.objective,
                verbose=self.verbose,
            )
        else:
            raise NotImplementedError(
                "Only Exhaustive, Halving and Optuna are implemented."
            )
        # Get results
        results = grid_search.fit(self.x[self.feature_sets[feature_set]], self.y)
        results = results.sort_values("worst_case", ascending=False)

        # Warn when best hyperparameters are close to predefined grid
        edge_params = grid_search.get_parameter_min_max()
        best_params = pd.Series(results["params"].iloc[0], name="best")
        params = edge_params.join(best_params, how="inner")

        def warn_when_too_close_to_edge(param: pd.Series, tol=0.01):
            # Min-max scaling
            scaled = np.array(param["best"]) / (param["max"] - param["min"])
            min_, best, max_ = param["min"], param["best"], param["max"]
            scaled = (best - min_) / (max_ - min_)
            # Check if too close and warn if so
            if not (tol < scaled < 1 - tol):
                msg = "Optimal value for parameter is very close to edge case: "
                if min(abs(min_), abs(max_)) < 1:
                    msg += f"{param.name}={best:.2e} (range: {min_:.2e}...{max_:.2e})"
                else:
                    msg += f"{param.name}={best} (range: {min_}...{max_})"
                warnings.warn(msg)

        params.apply(warn_when_too_close_to_edge, axis=1)

        return results

    def _get_main_predictors(self, data):
        """
        Using Shapely Additive Explanations, this function calculates the main
        predictors for a given prediction and sets them into the class' memory.
        """
        # Shap is not implemented for all models.
        if type(self.best_model).__name__ in [
            "SVC",
            "BaggingClassifier",
            "RidgeClassifier",
            "LinearRegression",
            "LogisticRegression",
            "SVR",
            "BaggingRegressor",
        ]:
            fi = self.settings["feature_processing"]["feature_importance_"]
            features = list(fi["shap"].keys())
            values = list(fi["shap"].values())
            self._main_predictors = {
                features[i]: values[i] for i in range(len(features))
            }

        else:
            if type(self.best_model).__module__[:5] == "amplo":
                shap_values = np.array(
                    TreeExplainer(self.best_model.model).shap_values(data)
                )
            else:
                shap_values = np.array(TreeExplainer(self.best_model).shap_values(data))

            # Shape them (multiclass outputs ndim=3, for binary/regression ndim=2)
            if shap_values.ndim == 3:
                shap_values = shap_values[1]

            # Take mean over samples
            shap_values = np.mean(shap_values, axis=0)

            # Sort them
            inds = sorted(range(len(shap_values)), key=lambda x: -abs(shap_values[x]))

            # Set class attribute
            self._main_predictors = dict(
                [(data.keys()[i], float(abs(shap_values[i]))) for i in inds]
            )

        return self._main_predictors



#  Copyright (c) 2022 by Amplo.

import copy
import math
import os
from datetime import datetime
from typing import TYPE_CHECKING
from urllib.error import URLError

import matplotlib.pyplot as plt
import numpy as np
from fpdf import FPDF
from sklearn.metrics import RocCurveDisplay, auc
from sklearn.model_selection import StratifiedKFold

from amplo.automl.feature_processing import (
    NopFeatureExtractor,
    StaticFeatureExtractor,
    TemporalFeatureExtractor,
)
from amplo.utils.logging import logger
from amplo.utils.util import deprecated

if TYPE_CHECKING:
    from amplo import Pipeline


@deprecated(
    "This is no longer used, all information is available in the created settings.json."
)
class BinaryDocumenting(FPDF):
    def __init__(self, pipeline: "Pipeline"):
        super().__init__()

        # Settings
        self.WIDTH = 210
        self.HEIGHT = 297
        self.lh = 6  # Line Height
        self.pm = 15  # Paper margin
        self.tm = 8  # Text Margins
        self.set_margins(self.pm, self.pm, self.pm)

        # Args
        self.p = pipeline
        self.name = pipeline.name

        # Initiates
        self.model = None
        self.mName = None
        self.feature_set = None
        self.metrics = None
        self.outputMetrics = None
        self.confusion_matrix = None
        self.cv = None
        self.x = None
        self.y = None
        self.featureImportances = {}

    def header(self):
        try:  # No problem if not possible due lack of internet
            self.image(
                "https://raw.githubusercontent.com/nielsuit227/AutoML/main/Amplo/Static/logo.png",
                x=self.pm,
                y=self.pm,
                w=55,
                h=15,
            )
        except URLError:
            pass
        self.set_font("Helvetica", "", 14)
        self.set_text_color(100, 100, 100)
        self.cell(self.WIDTH - 80)
        self.cell(0, 15, "Amplo AutoML Documentation", 0, 0, align="R")
        self.set_font("Helvetica", "", 10)
        self.ln(6)
        self.cell(
            0,
            15,
            datetime.now().strftime("%d %b %Y - %H:%M"),
            0,
            0,
            align="R",
        )
        self.ln(25)

    def create(self, model, feature_set):
        """
        Creates the entire documentation.

        Parameters
        ----------
        model : Model to document / validate
        feature_set : Feature set to document / validate on
        """
        # Asserts
        assert model is not None, "Model cannot be none."
        assert feature_set in self.p.feature_sets.keys(), "Feature set unavailable"

        # Set variables
        self.model = model
        self.mName = type(model).__name__
        self.feature_set = feature_set

        # Analyse the model
        self.prepare_data()
        self.analyse()

        # Check if folder needs creation
        path = self.p.main_dir + "Documentation/v{}/{}_{}.pdf".format(
            self.p.version, self.mName, feature_set
        )
        if not os.path.exists(path[: path.rfind("/")]):
            os.makedirs(path[: path.rfind("/")])

        # Create PDF
        self.add_page()
        self.add_h1("{}".format(self.name))
        self.add_h2("{} - v{}".format(self.mName, self.p.version))
        self.add_text(self.model_description)
        self.model_performance()
        self.validation()
        self.parameters()
        self.features()
        self.data()
        self.score_board()
        self.output(
            self.p.main_dir
            + "Documentation/v{}/{}_{}.pdf".format(
                self.p.version, self.mName, self.feature_set
            )
        )

    def prepare_data(self):
        """
        Although data of the Pipeline (self.p.x and self.p.y), is already cleaned, it
        is not normalized or sequenced.
        @return:
        @rtype:
        """
        x, y = self.p.x, self.p.y
        self.x, self.y = x.to_numpy(), y.to_numpy()

    def analyse(self):
        # Metrics & Confusion Matrix
        accuracy = []
        precision = []
        sensitivity = []
        specificity = []
        f1_score = []
        area_under_curves = []
        true_positive_rates = []
        cm = np.zeros((self.p.cv_splits, 2, 2))
        mean_fpr = np.linspace(0, 1, 100)

        # Plot Initiator
        fig, ax = plt.subplots(
            math.ceil(self.p.cv_splits / 2),
            2,
            sharex="all",
            sharey="all",
            figsize=[12, 8],
        )
        fig.suptitle(
            "{}-Fold Cross Validated Predictions - {} ({})".format(
                self.p.cv_splits, self.mName, self.feature_set
            )
        )
        fig2, ax2 = plt.subplots(figsize=[12, 8])

        # Modelling
        self.cv = StratifiedKFold(n_splits=self.p.cv_splits, shuffle=self.p.shuffle)
        for i, (t, v) in enumerate(self.cv.split(self.x, self.y)):
            n = len(v)
            xt, xv, yt, yv = (
                self.x[t],
                self.x[v],
                self.y[t].reshape((-1)),
                self.y[v].reshape((-1)),
            )
            model = copy.deepcopy(self.model)
            model.fit(xt, yt)
            predictions = model.predict(xv).reshape((-1))

            # Metrics
            tp = np.logical_and(np.sign(predictions) == 1, yv == 1).sum()
            tn = np.logical_and(np.sign(predictions) == 0, yv == 0).sum()
            fp = np.logical_and(np.sign(predictions) == 1, yv == 0).sum()
            fn = np.logical_and(np.sign(predictions) == 0, yv == 1).sum()
            accuracy.append((tp + tn) / n * 100)
            if tp + fp > 0:
                precision.append(tp / (tp + fp) * 100)
            if tp + fn > 0:
                sensitivity.append(tp / (tp + fn) * 100)
            if tn + fp > 0:
                specificity.append(tn / (tn + fp) * 100)
            if tp + fp > 0 and tp + fn > 0:
                f1_score.append(
                    2
                    * precision[-1]
                    * sensitivity[-1]
                    / (precision[-1] + sensitivity[-1])
                    if precision[-1] + sensitivity[-1] > 0
                    else 0
                )
            cm[i] = np.array([[tp, fp], [fn, tn]])

            # ROC calculations
            viz = RocCurveDisplay.from_predictions(
                yv, predictions, name="ROC fold {}".format(i + 1), ax=ax2
            )
            interp_tpr = np.interp(mean_fpr, viz.fpr, viz.tpr)
            interp_tpr[0] = 0.0
            true_positive_rates.append(interp_tpr)
            area_under_curves.append(viz.roc_auc)

            # Plot
            ax[i // 2][i % 2].plot(yv, c="#2369ec", alpha=0.6)
            ax[i // 2][i % 2].plot(predictions, c="#ffa62b")
            ax[i // 2][i % 2].set_title("Fold-{}".format(i))

        # Statistics on results
        totals = np.sum(cm, axis=(1, 2), keepdims=True)
        means = np.mean(cm / totals * 100, axis=0).astype("float")
        stds = np.std(cm / totals * 100, axis=0).astype("float")

        # Store
        self.metrics = {
            "Accuracy": [float(np.mean(accuracy)), float(np.std(accuracy))],
            "Precision": [float(np.mean(precision)), float(np.std(precision))],
            "Sensitivity": [float(np.mean(sensitivity)), float(np.std(sensitivity))],
            "Specificity": [float(np.mean(specificity)), float(np.std(specificity))],
            "F1 Score": [float(np.mean(f1_score)), float(np.std(f1_score))],
        }
        self.confusion_matrix = {
            "means": means,
            "stds": stds,
        }
        self.outputMetrics = copy.deepcopy(self.metrics)
        self.outputMetrics["True Positives"] = [
            means[0, 0],
            stds[0, 0],
        ]  # These are float32s
        self.outputMetrics["False Positives"] = [means[0, 1], stds[0, 1]]
        self.outputMetrics["False Negatives"] = [means[1, 0], stds[1, 0]]
        self.outputMetrics["True Negatives"] = [means[1, 1], stds[1, 1]]

        # Print Results
        logger.info(
            "[AutoML] Accuracy:        {:.2f} \u00B1 {:.2f} %".format(
                np.mean(accuracy), np.std(accuracy)
            )
        )
        logger.info(
            "[AutoML] Precision:       {:.2f} \u00B1 {:.2f} %".format(
                np.mean(precision), np.std(precision)
            )
        )
        logger.info(
            "[AutoML] Sensitivity:     {:.2f} \u00B1 {:.2f} %".format(
                np.mean(sensitivity), np.std(sensitivity)
            )
        )
        logger.info(
            "[AutoML] Specificity:     {:.2f} \u00B1 {:.2f} %".format(
                np.mean(specificity), np.std(specificity)
            )
        )
        logger.info(
            "[AutoML] F1-score:        {:.2f} \u00B1 {:.2f} %".format(
                np.mean(f1_score), np.std(f1_score)
            )
        )
        logger.info("[AutoML] Confusion Matrix:")
        logger.info("[AutoML] Prediction / true |     Faulty     |    Healthy      ")
        logger.info(
            "[AutoML]       Faulty      | {} |  {}".format(
                ("{:.1f} \u00B1 {:.1f} %".format(means[0, 0], stds[0, 0])).ljust(14),
                ("{:.1f} \u00B1 {:.1f} %".format(means[0, 1], stds[0, 1])).ljust(14),
            )
        )
        logger.info(
            "[AutoML]       Healthy     | {} |  {}".format(
                ("{:.1f} \u00B1 {:.1f} %".format(means[1, 0], stds[1, 0])).ljust(14),
                ("{:.1f} \u00B1 {:.1f} %".format(means[1, 1], stds[1, 1])).ljust(14),
            )
        )

        # Check whether plot is possible
        if type(self.model).__name__ == "Lasso" or "Ridge" in type(self.model).__name__:
            return

        # Adjust plots
        ax2.plot(
            [0, 1],
            [0, 1],
            linestyle="--",
            lw=2,
            color="#ffa62b",
            label="Chance",
            alpha=0.8,
        )
        mean_tpr = np.mean(true_positive_rates, axis=0)
        mean_tpr[-1] = 1.0
        mean_auc = auc(mean_fpr, mean_tpr)
        std_auc = np.std(area_under_curves)
        ax2.plot(
            mean_fpr,
            mean_tpr,
            color="#2369ec",
            label=r"Mean ROC (AUC = {:.2f} $\pm$ {:.2f})".format(mean_auc, std_auc),
            lw=2,
            alpha=0.8,
        )
        std_tpr = np.std(true_positive_rates, axis=0)
        true_pos_rates_upper = np.minimum(mean_tpr + std_tpr, 1)
        true_pos_rates_lower = np.maximum(mean_tpr - std_tpr, 0)
        ax2.fill_between(
            mean_fpr,
            true_pos_rates_lower,
            true_pos_rates_upper,
            color="#729ce9",
            alpha=0.2,
            label=r"$\pm$ 1 std. dev.",
        )
        ax2.set(
            xlim=[-0.05, 1.05],
            ylim=[-0.05, 1.05],
            title="ROC Curve - {}".format(self.mName),
        )
        ax2.legend(loc="lower right")
        if not os.path.exists(
            self.p.main_dir + "EDA/Validation/v{}/".format(self.p.version)
        ):
            os.makedirs(self.p.main_dir + "EDA/Validation/v{}/".format(self.p.version))
        roc_path = self.p.main_dir + "EDA/Validation/v{}/ROC_{}.png".format(
            self.p.version, self.mName
        )
        fig2.savefig(roc_path, format="png", dpi=200)
        cross_val_path = self.p.main_dir + "EDA/Validation/v{}/Cross_Val_{}.png".format(
            self.p.version, self.mName
        )
        fig.savefig(cross_val_path, format="png", dpi=200)

        # Feature Importance (only if EDA is not run)
        if not os.path.exists(
            self.p.main_dir + "EDA/Features/v{}/RF.png".format(self.p.version)
        ):
            if not os.path.exists(
                self.p.main_dir + "EDA/Features/v{}".format(self.p.version)
            ):
                os.makedirs(
                    self.p.main_dir + "EDA/Features/v{}/".format(self.p.version)
                )
            fig, ax = plt.subplots(figsize=[4, 6], constrained_layout=True)
            plt.subplots_adjust(left=0.5, top=1, bottom=0)
            ax.spines["right"].set_visible(False)
            ax.spines["bottom"].set_visible(False)
            ax.spines["top"].set_visible(False)
            fi = self.p.feature_processor.feature_importance_["rf"]
            plt.barh(list(fi)[:15], width=list(fi.values())[:15], color="#2369ec")
            fig.savefig(
                self.p.main_dir + "EDA/Features/v{}/RF.png".format(self.p.version),
                format="png",
                dpi=200,
            )

    def check_new_page(self, margin=220):
        if self.get_y() > margin:
            self.add_page()
            return True
        else:
            return False

    def add_h1(self, title):
        self.set_font("Helvetica", "B", 18)
        self.set_text_color(80, 80, 80)
        self.cell(0, 0, title, 0, 0)
        self.ln(self.lh * 2)

    def add_h2(self, title):
        self.set_font("Helvetica", "B", 16)
        self.set_text_color(80, 80, 80)
        self.cell(0, 0, title, 0, 0)
        self.ln(self.lh)

    def add_h3(self, title, **args):
        self.set_font("Helvetica", "B", 14)
        self.set_text_color(80, 80, 80)
        self.cell(0, 0, title, **args)
        self.ln(self.lh)

    def add_text(self, text):
        self.set_font("Helvetica", "", 12)
        self.set_text_color(80, 80, 80)
        self.multi_cell(0, 5, text)

    def add_image(self, path, height, width):
        self.image(path, h=height, w=width, x=self.WIDTH - width)

    @property
    def model_description(self):
        if "CatBoost" in self.mName:
            return (
                "CatBoost, or Categorical Boosting,  is an algorithm for gradient "
                "boosting on decision trees, with natural implementation for "
                "categorical variables. It is similar to XGBoost and LightGBM but "
                "differs in implementation of the optimization algorithm. We often "
                "see this algorithm performing very well."
            )
        elif "XGB" in self.mName:
            return (
                "XGBoost, or Extreme Gradient Boosting, is an algorithm for gradient "
                "boosting on decision trees. It trains many decision trees sequentially"
                ", the additional tree always trying to mitigate the error of the whole"
                " model. XGBoost was the first gradient boosting algorithm to be "
                "implemented and is currently widely adopted in the ML world. "
            )
        elif "LGBM" in self.mName:
            return (
                "LightGBM, or Light Gradient Boosting Machine, is an iteration on the "
                "XGBoost algorithm. Similarly, it uses gradient boosting with decision "
                "trees. However, XGBoost tend to be slow for a larger number of samples"
                " (>10.000), but with leaf-wise growth instead of depth-wise growth, "
                "LightGBM increases training speed significantly. Performance is often "
                "close to XGBoost."
            )
        elif "HistG" in self.mName:
            return (
                "SciKits implementation of LightGBM, or Light Gradient Boosting Machine"
                ", is an iteration on the XGBoost algorithm. Similarly, uses gradient"
                " boosting with decision trees. However, XGBoost tend to be slow for a"
                " larger number of samples (>10.000), but with a different growth "
                "algorithm, LightGBM increases training speed significantly."
                " Performance is often close to XGBoost."
            )
        elif "GradientB" in self.mName:
            return (
                "SciKits implementation of XGBoost, or Extreme Gradient Boosting, is an"
                " algorithm for gradient boosting on decision trees. It trains many "
                "decision trees sequentially, the additional tree always trying to "
                "mitigate the error of the whole model. XGBoost was the first gradient"
                " boosting algorithm to be implemented and is currently widely adopted"
                " in the ML world. "
            )
        elif "RandomForest" in self.mName:
            return (
                "Random Forest, is an ensemble algorithm that combines many (100-1000) decision trees and "
                "predicts the average of all trained trees. Though gradient boosting methods often outperform "
                "Random Forest, some data characteristics favor the Random Forests performance. "
            )
        elif "Linear" in self.mName:
            return (
                "Linear models are simple algorithms where the inputs are multiplied by optimized weights to "
                "predict the output. "
            )
        elif "Bagging" in self.mName:
            return (
                "Bagging algorithms is an ensemble algorithm. Just like a Random "
                "Forest, it trains many Decision Trees. It then makes a prediction "
                "based on a voting base, the average of the prediction of the "
                "individual Decision Trees will be predicted by the Bagging algorithm."
                "Contrary to a Random Forest, the Bagging algorithm does not allocate "
                "subsets of features or data to the individual Decision Trees. "
            )
        elif "Stacking" in self.mName:
            return (
                "A Stacking algorithm is a linear algorithm that uses the prediction of"
                " various models. In our AutoML pipeline, we first train all included "
                "models, optimize the hyper parameters of the well performing ones and "
                "then take the three best combinations. Additionally, we add a Naive "
                "Bayes, a Linear Model and a K-Nearest Neighbors algorithm on top."
            )
        else:
            return """Model description yet to be included in this documenter."""

    def model_performance(self):
        if not self.check_new_page():
            self.ln(self.lh)
        self.add_h2("Model Performance")
        self.add_text(
            "Model performance is analysed by various metrics. This model has been selected based on the {} "
            "score.".format(self.p.objective)
        )

        # Metrics
        self.set_font("Helvetica", "B", 12)
        self.ln(self.lh)
        self.cell(w=50, h=self.lh, txt="Metric", border="B", align="C")
        self.cell(w=50, h=self.lh, txt="Score", border="LB", align="C")
        self.set_font("Helvetica", "", 12)
        for k, v in self.metrics.items():
            self.ln(self.lh)
            self.cell(w=50, h=self.lh, txt=k, border="R", align="L")
            self.cell(
                w=50,
                h=self.lh,
                txt="{:.2f} \u00B1 {:.2f} %".format(v[0], v[1]),
                border="L",
                align="C",
            )
        self.ln(self.lh * 3)

        # Confusion Matrix
        n_classes = 2
        cell_width = int(
            (self.WIDTH - self.pm * 4 - self.tm * n_classes) / (n_classes + 2)
        )
        self.add_h3("Confusion Matrix")

        # First row
        self.set_font("Helvetica", "B", 12)
        self.cell(w=cell_width * 2, h=self.lh, txt="", align="L", border="R")
        self.cell(w=cell_width * n_classes, h=self.lh, txt="True Class", align="C")
        self.ln(self.lh)

        # Second Row
        self.cell(w=cell_width * 2, h=self.lh, txt="", align="L", border="BR")
        self.cell(w=cell_width, h=self.lh, txt="Faulty", align="C", border="B")
        self.cell(w=cell_width, h=self.lh, txt="Healthy", align="C", border="B")
        self.ln(self.lh)

        # Third Row (first with values)
        self.cell(w=cell_width, h=self.lh * 2, txt="Prediction", align="L")
        self.cell(w=cell_width, h=self.lh, txt="Faulty", align="L", border="R")
        self.set_font("Helvetica", "", 12)
        self.cell(
            w=cell_width,
            h=self.lh,
            txt="{:.2f} \u00B1 {:.2f} %".format(
                self.confusion_matrix["means"][0][0],
                self.confusion_matrix["stds"][0][0],
            ),
            align="C",
        )
        self.cell(
            w=cell_width,
            h=self.lh,
            txt="{:.2f} \u00B1 {:.2f} %".format(
                self.confusion_matrix["means"][0][1],
                self.confusion_matrix["stds"][0][1],
            ),
            align="C",
        )
        self.ln(self.lh)

        # Fourth Row
        self.set_font("Helvetica", "B", 12)
        self.cell(w=cell_width, h=self.lh, txt="", align="L")
        self.cell(w=cell_width, h=self.lh, txt="Healthy", align="L", border="R")
        self.set_font("Helvetica", "", 12)
        self.cell(
            w=cell_width,
            h=self.lh,
            txt="{:.2f} \u00B1 {:.2f} %".format(
                self.confusion_matrix["means"][1][0],
                self.confusion_matrix["stds"][1][0],
            ),
            align="C",
        )
        self.cell(
            w=cell_width,
            h=self.lh,
            txt="{:.2f} \u00B1 {:.2f} %".format(
                self.confusion_matrix["means"][1][1],
                self.confusion_matrix["stds"][1][1],
            ),
            align="C",
        )
        self.ln(self.lh)

    def validation(self):
        if not self.check_new_page():
            self.ln(self.lh)
        self.ln(self.lh * 2)
        self.add_h3("Area Under Curve & Cross Validation Plots")
        x, y = self.get_x(), self.get_y()
        path = self.p.main_dir + "EDA/Validation/v{}/ROC_{}.png".format(
            self.p.version, self.mName
        )
        self.image(x=x, y=y, w=90, h=60, name=path)
        path = path[: path.rfind("/")] + "/Cross_Val_{}.png".format(self.mName)
        self.image(x=x + self.WIDTH / 2 - self.pm, y=y, w=90, h=60, name=path)
        self.ln(self.lh)
        self.add_page()
        self.add_h3("Validation Strategy")
        self.add_text(
            "All experiments are cross-validated. This means that every time a model's "
            "performance is evaluated, it's trained on one part of the data, and test on another. Therefore, "
            "the model is always test against data it has not yet been trained for. This gives the best "
            "approximation for real world (out of sample) performance. The current validation strategy used "
            "is {}, with {} splits and {} shuffling the data.".format(
                type(self.cv).__name__,
                self.p.cv_splits,
                "with" if self.p.shuffle else "without",
            )
        )
        self.ln(self.lh)

    def parameters(self):
        if not self.check_new_page():
            self.ln(self.lh)
        params = self.model.get_params()
        n_params = len(params)
        keys, values = list(params.keys()), list(params.values())
        self.ln(self.lh)
        self.add_h2("Model Parameters")
        self.set_font("Helvetica", "B", 12)

        # Double rows
        if n_params > 15:
            # First row
            w = 30
            self.cell(w=w + 30, h=self.lh, txt="Parameter", align="C", border="RB")
            self.cell(w=w, h=self.lh, txt="Value", align="C", border="B")
            self.cell(w=5, h=self.lh, txt="")
            self.cell(w=w + 30, h=self.lh, txt="Parameter", align="C", border="RB")
            self.cell(w=w, h=self.lh, txt="Value", align="C", border="B")
            self.set_font("Helvetica", "", 12)
            n_rows = -(-n_params // 2)
            for i in range(n_rows):
                self.ln(self.lh)
                self.cell(w=w + 30, h=self.lh, txt=keys[i * 2], align="L", border="R")
                value = (
                    "{:.4e}".format(values[i * 2])
                    if isinstance(values[i * 2], float)
                    else str(values[i * 2])
                )
                self.cell(w=w, h=self.lh, txt=value, align="C")
                self.cell(w=5, h=self.lh),
                if i * 2 + 1 < n_params:
                    self.cell(
                        w=w + 30, h=self.lh, txt=keys[i * 2 + 1], align="L", border="R"
                    )
                    value = (
                        "{:.4e}".format(values[i * 2 + 1])
                        if isinstance(values[i * 2 + 1], float)
                        else str(values[i * 2 + 1])
                    )
                    self.cell(w=w, h=self.lh, txt=value, align="C")

        # Single rows
        else:
            w = 50
            self.cell(w=w, h=self.lh, txt="Parameter", align="L", border="RB")
            self.cell(w=w, h=self.lh, txt="Value", align="L", border="B")
            for i in range(n_params):
                self.ln(self.lh)
                self.cell(w=w, h=self.lh, txt=keys[i], align="L", border="R")
                value = (
                    "{:.4e}".format(values[i])
                    if isinstance(values[i], float)
                    else str(values[i])
                )
                self.cell(w=w, h=self.lh, txt=value, align="R")

    def features(self):
        feature_extractor = self.p.feature_processor.feature_extractor
        if isinstance(feature_extractor, NopFeatureExtractor):
            features = {
                "Raw Features": feature_extractor.features_,
            }
        elif isinstance(feature_extractor, StaticFeatureExtractor):
            features = {
                "Raw Features": feature_extractor.raw_features_,
                "Arithmetic Features": feature_extractor.cross_features_,
                "K-Means Features": feature_extractor.k_means_features_,
                "Trigonometric Features": feature_extractor.trigo_features_,
                "Inverse Features": feature_extractor.inverse_features_,
            }
        elif isinstance(feature_extractor, TemporalFeatureExtractor):
            features = {
                "Raw Features": feature_extractor.raw_features_,
                "Wavelet Features": feature_extractor.wav_features_,
            }
        else:
            msg = (
                f"Feature extractor of type "
                f"{type(feature_extractor).__name__} not recognized."
            )
            raise TypeError(msg)
        features.update(
            {
                "Collinear Features": self.p.feature_processor.collinear_cols_,
            }
        )

        if not self.check_new_page():
            self.ln(20)
        self.add_h2("Features")

        # Feature Extraction
        self.add_h3("Feature Extraction")
        self.add_text(
            "First, features that are collinear (a * x = y) up to {} % were removed. "
            "This resulted in {} removed features: {}{}".format(
                self.p.information_threshold * 100,
                len(features["Collinear Features"]),
                ", ".join([i for i in features["Collinear Features"][:20]]),
                ", ..." if len(features["Collinear Features"]) > 20 else " ",
            )
        )
        self.check_new_page(margin=220)
        self.add_text(
            "Subsequently, the features were manipulated and analysed to extract "
            "additional information. All promising combinations are analysed with a "
            "single shallow decision tree."
        )
        features.pop("Collinear Features")
        self.ln()
        self.set_font("Helvetica", "B", 12)
        self.cell(w=50, h=self.lh, txt="Sort Feature", border="BR", align="C")
        self.cell(w=50, h=self.lh, txt="Quantity", border="B", align="C")
        # Not supported now as multi cell adds whitespace in between rows, maybe appendix?
        # self.cell(w=120, h=self.lh, txt='Features', border='B', align='C')
        self.set_font("Helvetica", "", 12)
        # n_rows = max([len(v) for v in features.values()])
        for k, v in features.items():
            self.ln(self.lh)
            # h = max(- (- sum([len(i) for i in v]) // 60), 1)
            self.cell(w=50, h=self.lh, txt=k, border="R", align="L")
            self.cell(w=50, h=self.lh, txt="{}".format(len(v)), align="C")
            # self.multi_cell(w=140, h=self.lh, txt=', '.join([i for i in v]), align='L')
        self.check_new_page()

        # Feature Selection
        self.add_h3("Feature Selection")
        self.set_font("Helvetica", "", 12)
        y = self.get_y()
        self.multi_cell(
            w=80,
            h=self.lh,
            txt="Using a Random Forest model, the non-linear Feature Importance is analysed. The feature "
            "importance is measured in Mean Decrease in Gini Impurity. The feature importance is used "
            "to create two feature sets, one that contains 85% of all feature importance, and one that "
            "contains all features that contribute more than 1% to the total feature importance.\nThe "
            "top 15 feature with their mean decrease in Gini impurity are visualized on the right.",
        )
        path = self.p.main_dir + "EDA/Features/v{}/RF.png".format(self.p.version)
        if os.path.exists(path):
            self.image(name=path, x=110 + self.pm, y=y - 10, w=80, h=120)

    def data(self):
        if not self.check_new_page():
            self.ln(30)
        self.add_h2("Data Processing")
        self.add_text(
            "The following data manipulations were made to clean the data:\n"
            "  1.  Removed {} duplicate columns and {} duplicate rows\n"
            "  2.  Removed {} outliers with {}\n"
            "  3.  Imputed {} missing values with {}\n"
            "  4.  Removed {} columns with constant values\n".format(
                self.p.data_processor.removedDuplicateColumns,
                self.p.data_processor.removedDuplicateRows,
                self.p.data_processor.removedOutliers,
                self.p.data_processor.outlier_removal,
                self.p.data_processor.imputedMissingValues,
                self.p.data_processor.missing_values,
                self.p.data_processor.removedConstantColumns,
            )
        )

    def score_board(self):
        if self.p.results is None:
            return
        if not self.check_new_page():
            self.ln(self.lh)
        scores = self.p.sort_results(self.p.results)
        self.ln(self.lh)
        self.add_h2("Model Score Board")
        self.add_text(
            "Not only the {} has been optimized by the AutoML pipeline. In total, {} models where trained. "
            "The following table shows the performance of the top 10 performing models:".format(
                self.mName, len(self.p.results)
            )
        )
        self.ln(self.lh)
        # Table
        self.set_font("Helvetica", "B", 12)
        self.cell(w=70, h=self.lh, txt="Model", border="RB", align="C")
        self.cell(w=70, h=self.lh, txt=self.p.objective, border="B", align="C")
        # no space for parameters, move to appendix if wanted
        # self.cell(w=100, h=self.lh, txt='Parameters', border='B', align='C')
        self.set_font("Helvetica", "", 12)
        for i in range(min(10, len(scores))):
            self.ln(self.lh)
            self.cell(
                w=70,
                h=self.lh,
                txt="{}".format(scores.iloc[i]["model"]),
                border="R",
                align="C",
            )
            self.cell(
                w=70,
                h=self.lh,
                txt="{:.4f} \u00B1 {:.4f} %".format(
                    scores.iloc[i]["mean_objective"], scores.iloc[i]["std_objective"]
                ),
                align="C",
            )



#  Copyright (c) 2022 by Amplo.

from copy import deepcopy


class PartialBoostingRegressor:
    """
    Wrapper for boosting models which limits the number of estimators being used in
    the prediction.
    """

    _estimator_type = "regressor"

    _SUPPORTED_MODELS = (
        "AdaBoostRegressor",
        "GradientBoostingRegressor",
        "LGBMRegressor",
        "XGBRegressor",
        "CatBoostRegressor",
    )

    def __init__(self, model, step):
        """
        Construct wrapper for model with `predict` and `predict_proba` methods.

        Parameters
        ----------
        model
            boosting model to wrap.
        step : int
            Number of iterations/estimators to limit the model on predictions.
        """
        self.model_class = type(model).__name__
        self.step = step
        if self.model_class in ["AdaBoostRegressor", "GradientBoostingRegressor"]:
            self.model = deepcopy(model)
            self.model.estimators_ = self.model.estimators_[: self.step]
        else:
            self.model = model

    def predict(self, x):
        if self.model_class in [
            "AdaBoostRegressor",
            "GradientBoostingRegressor",
        ]:
            return self.model.predict(x)
        elif self.model_class == "LGBMRegressor":
            return self.model.predict(x, num_iteration=self.step)
        elif self.model_class == "XGBRegressor":
            return self.model.predict(x, iteration_range=(0, self.step))
        elif self.model_class == "CatBoostRegressor":
            return self.model.predict(x, ntree_end=self.step)
        else:
            raise ValueError("Incorrect model type.")

    @classmethod
    def n_estimators(cls, model):
        model_class = type(model).__name__
        if model_class in [
            "AdaBoostRegressor",
            "GradientBoostingRegressor",
        ]:
            return len(model.estimators_)
        elif model_class == "LGBMRegressor":
            return model.num_trees()
        elif model_class == "XGBRegressor":
            return model.model.num_boosted_rounds()
        elif model_class == "CatBoostRegressor":
            return model.model.tree_count_
        else:
            raise ValueError("Incorrect model type.")



#  Copyright (c) 2022 by Amplo.

"""
Feature processor for extracting temporal features.
"""
import re
from warnings import warn

import numpy as np
import pandas as pd
import pywt
from scipy import signal, stats

from amplo.automl.feature_processing._base import (
    BaseFeatureExtractor,
    sanitize_dataframe,
)
from amplo.base.exceptions import ExperimentalWarning
from amplo.utils import check_dtypes

__all__ = ["TemporalFeatureExtractor", "pool_single_index", "pool_multi_index"]


def _n_mean_crossings(x):
    """
    Calculates the number of crossings of x on mean.

    A crossing is defined as two sequential values where the first value is lower than
    mean and the next is greater, or vice-versa.

    Parameters
    ----------
    x : np.ndarray or pd.Series
        Time series data to calculate the feature of.

    Returns
    -------
    float
        Number of mean crossings.
    """
    # Reference: https://stackoverflow.com/questions/3843017/efficiently-detect-sign-changes-in-python  # noqa: E501
    positive = np.signbit(x - np.mean(x))
    return np.where(np.bitwise_xor(positive[1:], positive[:-1]))[0].size


def _cid_ce(x, normalize=True):
    """
    Calculates an estimate for a time series complexity.

    Parameters
    ----------
    x : np.ndarray or pd.Series
        Time series data to calculate the feature of.
    normalize : bool
        Whether to z-transform the time series.

    Returns
    -------
    float
        Estimated complexity.
    """
    # Reference: https://tsfresh.readthedocs.io/en/latest/api/tsfresh.feature_extraction.html  # noqa: E501
    if normalize:
        s = np.std(x)
        if s != 0:
            x = (x - np.mean(x)) / s
        else:
            return 0.0

    x = np.diff(x)
    return np.sqrt(np.dot(x, x))


def pool_single_index(data, ws, agg_func):
    """
    Pools series data with given aggregation function.

    Parameters
    ----------
    data : pd.Series
        Data to be pooled.
    ws : int
        Window size; an unsigned integer.
    agg_func : typing.Callable or typing.Dict[str, typing.Callable]
        Function or dictionary of functions to be called for aggregation.
        cf. Examples

    Returns
    -------
    pd.DataFrame
        Pooled data.

    Examples
    --------
    >>> pool_single_index(data, 10, np.mean)
    >>> pool_single_index(data, 10, {"min": np.min, "max": np.max})

    Notes
    -----
    This function assumes the data to be of one time-series.
    """
    # Fill tail when number of missing values in the tail is greater than half
    # the window size.  Otherwise, remove the tail.
    tail = data.shape[0] % ws
    n_missing_in_tail = ws - tail
    if 0 < n_missing_in_tail < ws / 2:
        add_to_tail = data.iloc[-n_missing_in_tail:]
        data = pd.concat([data, add_to_tail])
    elif tail != 0:
        data = data.iloc[:-tail]

    # Assure that agg_func is a dictionary
    if not isinstance(agg_func, dict):
        agg_func = {agg_func.__name__: agg_func}

    # Pooling
    np_data = data.values.reshape(-1, 1, ws)  # noqa
    pooled_data = pd.DataFrame(index=data.index[::ws])
    with np.errstate(divide="ignore", invalid="ignore"):  # ignore true_divide warnings
        for key, func in agg_func.items():
            try:
                pooled_data[key] = func(np_data, axis=-1)
            except TypeError:
                # func doesn't support an axis keyword argument
                # Note: This is a very lot slower but at least gives the correct output.
                pooled_data[key] = np.apply_along_axis(func, axis=-1, arr=np_data)

    return pooled_data


def pool_multi_index(data, ws, agg_func):
    """
    Groupby-pools series data with given aggregation function.

    Parameters
    ----------
    data : pd.Series
        Data to be pooled.
    ws : int
        Window size; an unsigned integer.
    agg_func : typing.Callable or typing.Dict[str, typing.Callable]
        Function or dictionary of functions to be called for aggregation.
        cf. Examples

    Returns
    -------
    pd.Series or pd.DataFrame
        Pooled data.

    Examples
    --------
    # returns pd.Series
    >>> pool_multi_index(data, 10, np.mean)
    # returns pd.DataFrame
    >>> pool_multi_index(data, 10, {"min": np.min, "max": np.max})

    Notes
    -----
    This function makes use of the multi-index.
    """

    pooled_data = data.groupby(level=0).apply(
        lambda group: pool_single_index(group.droplevel(0), ws, agg_func)
    )

    return pooled_data


def _extract_wavelets(series, scales, wavelet, name=None):
    assert isinstance(series, pd.Series)

    # Transform
    # Note that sampling_frequency does not matter.
    coeffs, _ = pywt.cwt(series, scales=scales, wavelet=wavelet)

    # Make dataframe
    columns = [f"{name or series.name}__wav__{wavelet}__{s}" for s in scales]
    x_out = pd.DataFrame(coeffs.real.T, index=series.index, columns=columns)

    return x_out


class TemporalFeatureExtractor(BaseFeatureExtractor):
    """
    Feature extractor for temporal data.

    Parameters
    ----------
    mode : str
        Model mode: {"classification", "regression"}.
    fit_wavelets : typing.List of str, optional
        Wavelet names for feature exploration (for fitting only).
        Defaults: ["cmor1.5-1.0", "gaus4", "gaus7", "cgau2", "cgau6", "mexh"]
    verbose : int
        Verbosity for logger.

    Notes
    -----
    Valid ``wavelet`` parameters can be found via:
        >>> import pywt
        >>> pywt.wavelist()
    """

    _add_to_settings = ["window_size_", *BaseFeatureExtractor._add_to_settings]

    def __init__(self, mode="notset", fit_wavelets=None, verbose=0):
        warn(
            "TemporalFeatureExtractor is an experimental feature.", ExperimentalWarning
        )
        super().__init__(mode=mode, verbose=verbose)

        if self.mode == "regression":
            # Some notes for implementing regression:
            #  - It does not make sense to pool features and target.
            #  - Wavelet transformations may still add some value.
            msg = (
                "TemporalFeatureExtractor is not ready for regression. "
                "Behavior probably won't meet your expectations!"
            )
            warn(msg, UserWarning)

        if fit_wavelets is None:
            # Set defaults
            fit_wavelets = ["cmor1.5-1.0", "gaus4", "gaus7", "cgau2", "cgau6", "mexh"]
        else:
            check_dtypes([("fit_wavelets", fit_wavelets, list)])
            check_dtypes([("fit_wavelets_item", item, str) for item in fit_wavelets])
        self.fit_wavelets = fit_wavelets

    def _fit_transform(self, x, y=None, **fit_params):
        self.logger.info("Start fitting data.")

        # Input checks
        x, y = self._check_x_y(x, y)
        numeric_cols = [
            col for col, typ in zip(x, x.dtypes) if np.issubdtype(typ, np.number)
        ]
        if set(numeric_cols) != set(x):
            warn(
                "Handling non-numeric data is (currently) not supported. "
                "Corresponding columns will be ignored.",
                UserWarning,
            )
            x = x[numeric_cols]

        # Initialize fitting
        self._set_validation_model()
        self._set_window_size(x.index)

        # Calculate baseline scores (w/o taking time into account)
        x_sampled = x.sample(n=min(10_000, y.shape[0]), random_state=93837)
        y_sampled = y.loc[x_sampled.index]
        self._init_feature_baseline_scores(x_sampled, y_sampled)
        del x_sampled, y_sampled

        # Fit features
        y_pooled = self._pool_target(y)
        x_out = pd.concat(
            [
                self._fit_transform_raw_features(x, y_pooled, update_baseline=True),
                self._fit_transform_wav_features(x, y_pooled, update_baseline=True),
            ],
            axis=1,
        )

        self.logger.info("Finished fitting.")
        return sanitize_dataframe(x_out[self.features_])

    def _transform(self, x, y=None):
        self.logger.info("Transforming data.")

        # Handle input
        x = self._check_x(x, convert_single_index=True)

        # Apply transformations
        x_out = pd.concat(
            [
                self._transform_raw_features(x),
                self._transform_wav_features(x),
            ],
            axis=1,
        )

        return sanitize_dataframe(x_out[self.features_])

    # ----------------------------------------------------------------------
    # Feature processing

    @property
    def raw_features_(self):
        out = [str(c) for c in self.features_ if not re.search(".+__.+__pool=.+", c)]
        return sorted(out)

    def _fit_transform_raw_features(self, x, y_pooled, update_baseline=True):
        self.logger.info("Fitting raw features.")

        # Pool all features
        x_pooled = self._pool_features(x, drop_nan_columns=True)

        # Score and decide which features to accept
        scores = self.select_scores(
            x_pooled.apply(self._calc_feature_scores, y=y_pooled, axis=0),
            best_n_per_class=50,
            update_baseline=update_baseline,
        )
        x_out = x_pooled[scores.columns]
        self.logger.info(f"Accepted {x_out.shape[1]} raw features.")

        # Add accepted features
        self.add_features(x_out)

        return x_out

    def _transform_raw_features(self, x):
        if not self.raw_features_:
            self.logger.debug("No raw features added.")
            dummy_y = pd.Series(np.zeros(len(x)), index=x.index, dtype=np.int32)
            return pd.DataFrame(index=self._pool_target(dummy_y).index)

        self.logger.info("Transforming raw features.")

        # Pooling
        pool_info = [tuple(c.split("__pool=")) for c in self.raw_features_]
        pool_info = pd.DataFrame(pool_info).groupby(0).agg(list)[1].to_dict()
        x_pooled = self._pool_features(x, pool_info)

        assert set(self.raw_features_) == set(
            x_pooled
        ), "Expected raw features do not match with actual."

        return x_pooled

    @property
    def wav_features_(self):
        out = [str(c) for c in self.features_ if re.search(".+__wav__.+", c)]
        return sorted(out)

    def _fit_transform_wav_features(self, x, y_pooled, update_baseline=True):
        self.logger.info("Fitting wavelet-transformed features.")

        # Initialize
        fs = 1.0  # correct sampling frequency only matters for plotting
        wavelets = self.fit_wavelets

        # Extract and score wavelet features for each column
        x_out = []
        all_scores = []
        for col in x:
            # Init extracted features
            col_feats = pd.DataFrame(index=y_pooled.index)

            # Get (local) peak frequencies of power spectral density
            freqs, pxx = signal.welch(x=x[col], fs=fs)
            peak_idx, _ = signal.find_peaks(np.log(pxx), prominence=0.3, distance=10)
            peak_freqs = freqs[peak_idx]

            for wv in wavelets:
                # Use the fact: scale = s2f_const / frequency.
                s2f_const = pywt.scale2frequency(wv, scale=1) * fs
                scales = np.round(s2f_const / peak_freqs, 2)

                # Extract and pool
                feats = (
                    x[col]
                    .groupby(level=0)
                    .apply(_extract_wavelets, scales=scales, wavelet=wv, name=col)
                )
                feats_pooled = self._pool_features(feats, drop_nan_columns=True)

                # Append to col_feats
                col_feats = pd.concat([col_feats, feats_pooled], axis=1)

            # Score and decide which features to accept
            scores = self.select_scores(
                col_feats.apply(self._calc_feature_scores, y=y_pooled, axis=0),
                best_n_per_class=50,
                update_baseline=update_baseline,
            )
            x_out += [col_feats[scores.columns]]
            all_scores += [scores]

        x_out = pd.concat(x_out, axis=1)
        all_scores = pd.concat(all_scores, axis=1)
        if update_baseline:
            self._update_feature_baseline_scores(all_scores)

        self.logger.info(f"Accepted {x_out.shape[1]} wavelet-transformed features.")

        # Add accepted features
        self.add_features(x_out)

        return x_out

    def _transform_wav_features(self, x):
        if not self.wav_features_:
            self.logger.debug("No wavelet-transformed features added.")
            dummy_y = pd.Series(np.zeros(len(x)), index=x.index, dtype=np.int32)
            return pd.DataFrame(index=self._pool_target(dummy_y).index)

        self.logger.info("Transforming wavelet-transformed features.")

        # Handle wavelet-transform features
        feat_info = []
        for f in self.wav_features_:
            col_name, intermediate = f.split("__wav__")
            wavelet, scale, _ = intermediate.split("__")
            feat_info.append((wavelet, col_name, scale))

        # Group the features by wavelets (moved to index)
        cols = ["wavelet", "col_name", "scales"]
        feat_by_wt = pd.DataFrame(feat_info, columns=cols).groupby("wavelet").agg(list)

        # Extract wavelets
        x_out = []
        for wv, info in feat_by_wt.iterrows():
            columns = sorted(set(info["col_name"]))
            scales = sorted(map(float, set(info["scales"])))

            # Transform and add to list
            for col in columns:
                x_out += [
                    x[col]
                    .groupby(level=0)
                    .apply(_extract_wavelets, scales=scales, wavelet=wv, name=col)
                ]
        x_out = pd.concat(x_out, axis=1)

        # Pooling
        pool_info = [tuple(c.split("__pool=")) for c in self.wav_features_]
        pool_info = pd.DataFrame(pool_info).groupby(0).agg(list)[1].to_dict()
        x_pooled = self._pool_features(x_out, pool_info)

        assert set(self.wav_features_) == set(
            x_pooled
        ), "Expected wavelet-transform features do not match with actual."

        return x_pooled

    # ----------------------------------------------------------------------
    # Utils

    @staticmethod
    def _check_x(x, copy=True, sanitize=True, convert_single_index=False):
        # Call parent checking method
        x_check = BaseFeatureExtractor._check_x(x, copy=copy, sanitize=sanitize)
        # Check multi-index
        n_index_cols = len(x.index.names)
        if n_index_cols == 1 and convert_single_index:
            x_check.index = pd.MultiIndex.from_product([[0], x_check.index])
        elif n_index_cols != 2:
            raise ValueError("Data is not properly multi-indexed.")
        return x_check

    def _set_window_size(self, index):
        # Count log sizes
        counts = pd.Series(index=index).fillna(0).groupby(level=0).count()

        # # We set the window size such that we can expect about 10 - 30 windows
        # # per log.  In the case where each log has the same size, we end up with
        # # 15 windows.
        # shortest, longest = counts.min(), counts.max()
        # window_size = int(shortest / 10 + longest / 30) // 2
        # self.window_size_ = max(1, window_size)

        if self.mode == "classification":
            # We set the window size such that it fits the smallest indices count.
            # However, we want that at least 100 data rows will be present after pool.
            ws = int(counts.min())
            if len(index) // ws < 100:
                ws = len(index) // 100

            # We don't want window sizes < 3. Pooling wouldn't make sense.
            if ws < 3:
                warn(
                    "The given data proposed using a window size smaller than 3 which "
                    "would be problematic. Setting it to 3 anyhow.",
                    UserWarning,
                )
                ws = 3

            self.window_size_ = ws

        elif self.mode == "regression":
            self.window_size_ = 1

        self.logger.debug(f"Set window size to {self.window_size_}.")

    def _pool_features(self, data, instruction=None, drop_nan_columns=False):
        """
        Pools data with given window size.

        Parameters
        ----------
        data : pd.DataFrame
            Data to be pooled.
        instruction : typing.Dict[str, list of str], optional
            Instructions for pooling. Each key corresponds the column name and
            its value defines the pooling names for the column.
            Default: Will calculate all implemented pools.
        drop_nan_columns : bool
            If false, all columns--no matter how many NaN values they have--will
            be returned. If true, columns with more than 10% of NaN entries will
            be removed.
            You are strongly encouraged to set this parameter only true when you
            are fitting data.

        Returns
        -------
        pd.DataFrame
            Pooled data.
        """
        # Define valid pooling functions.
        # Note: The `**k` in lambda expressions is to support `axis` keyword arguments.
        valid_pools = {
            # --- Basics ---
            "min": np.min,
            "max": np.max,
            "mean": np.mean,
            "std": np.std,
            "kurtosis": stats.kurtosis,
            "skew": stats.skew,
            # --- Characteristics ---
            "entropy": stats.entropy,
            "abs_energy": lambda x: np.dot(x, x),
            "abs_max": lambda x, **k: np.max(np.absolute(x), **k),
            # TODO: Add linear trend feature for `slope` and `stderror`. We have to
            #  adjust our pooling function to support target data as an input.
            #  c.f. tsfresh.feature_extraction.linear_trend
            # "linear_trend_slope": ...,
            # "linear_trend_stderror": ...,
            "n_mean_crossings": _n_mean_crossings,
            # --- Difference ---
            "abs_sum_of_changes": lambda x, **k: np.sum(np.abs(np.diff(x, **k)), **k),
            "mean_of_changes": lambda x, **k: np.mean(np.diff(x, **k), **k),
            "abs_mean_of_changes": lambda x, **k: np.mean(np.abs(np.diff(x, **k)), **k),
            "cid_ce": _cid_ce,
        }

        # Initialize
        if not instruction and self.mode == "regression":
            # Default: Use only mean feature (since window_size=1).
            instruction = {col: ["mean"] for col in data}
        elif not instruction:
            # Default: Apply all instructions on each column.
            instruction = {col: list(valid_pools) for col in data}
        else:
            # Validate
            requested_pools = []
            for col in instruction:
                requested_pools += set(instruction[col])
            invalid_pools = set(requested_pools) - set(valid_pools)
            if invalid_pools:
                msg = f"Invalid pooling names found: {invalid_pools}"
                raise ValueError(msg)

        # Pooling
        pooled_data = []
        for col, instr_lst in instruction.items():
            # Translate pooling instructions
            agg_func = {f"{col}__pool={name}": valid_pools[name] for name in instr_lst}
            # Apply
            pooled_col = pool_multi_index(data[col], self.window_size_, agg_func)
            pooled_data += [pooled_col]
        pooled_data = pd.concat(pooled_data, axis=1)

        # Sanitize
        if drop_nan_columns:
            rm_mask = pooled_data.isna().sum() > len(pooled_data) / 10
            pooled_data = pooled_data.loc[:, ~rm_mask]

        return sanitize_dataframe(pooled_data)

    def _pool_target(self, target):
        """
        Pools target data with given window size.

        Parameters
        ----------
        target : pd.Series
            Target data to be pooled.

        Returns
        -------
        pd.Series
            Pooled target data.
        """
        assert isinstance(target, pd.Series), "Data must be pandas.Series."
        assert isinstance(self.window_size_, int), "Invalid window size."

        dtype = target.dtype

        if self.mode == "classification":
            # This lambda function finds most occurring class per window.
            agg_func = lambda x: np.bincount(x, minlength=2).argmax()  # noqa: E731
        elif self.mode == "regression":
            agg_func = np.mean
        else:
            msg = f"Feature processor has an invalid mode: {self.mode}"
            raise AttributeError(msg)

        out = pool_multi_index(target, self.window_size_, agg_func)
        out = out.iloc[:, 0].astype(dtype).rename(target.name)

        return out



#  Copyright (c) 2022 by Amplo.

from copy import deepcopy

import lightgbm as lgb
import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error
from sklearn.model_selection import train_test_split


class LGBMRegressor:
    _estimator_type = "regressor"
    default_params = {
        "verbosity": -1,
        "force_col_wise": True,
        "num_boost_round": 1000,
        "objective": "rmse",
    }

    def __init__(self, **params):
        """
        Light GBM wrapper
        @param params: Model parameters

        callbacks [list[str]]: Callbacks as a string, transformed in a function here
        """
        self.params = None
        self.set_params(**params)
        self.model = None
        self.callbacks = []
        self.trained = False
        self.num_boost_round = 1000

    @staticmethod
    def convert_to_dataset(x, y=None):
        # Convert input
        assert type(x) in [
            pd.DataFrame,
            pd.Series,
            np.ndarray,
        ], "Unsupported data input format"
        if isinstance(x, np.ndarray) and len(x.shape) == 0:
            x = x.reshape((-1, 1))

        if y is None:
            return lgb.Dataset(x)

        else:
            assert type(y) in [pd.Series, np.ndarray], "Unsupported data label format"
            return lgb.Dataset(x, label=y)

    def fit(self, x, y):
        assert isinstance(x, np.ndarray) or isinstance(x, pd.DataFrame), (
            "X needs to be of type np.ndarray or " "pd.DataFrame"
        )
        assert isinstance(y, np.ndarray) or isinstance(
            y, pd.Series
        ), "Y needs to be of type np.ndarray or pd.Series"
        if isinstance(x, pd.DataFrame):
            x = x.to_numpy()
        if isinstance(y, pd.Series):
            y = y.to_numpy()

        # Split & Convert data
        train_x, test_x, train_y, test_y = train_test_split(
            x, y, test_size=0.1, shuffle=True
        )
        d_train = self.convert_to_dataset(train_x, train_y)
        d_test = self.convert_to_dataset(test_x, test_y)

        # Get params and split num_boost_round
        params = self.get_params()
        num_boost_round = params.pop("num_boost_round")

        # Model training
        self.model = lgb.train(
            params,
            d_train,
            num_boost_round=num_boost_round,
            valid_sets=[d_train, d_test],
            feval=self.eval_function,
            callbacks=[lgb.early_stopping(100, verbose=False)] + self._get_callbacks(),
        )
        self.trained = True

    def predict(self, x, *args, **kwargs):
        # todo check input data
        assert self.trained is True, "Model not yet trained"
        return self.model.predict(x, *args, **kwargs)

    def set_params(self, **params):
        # Add default
        for k, v in self.default_params.items():
            if k not in params:
                params[k] = v

        # Remove fit options
        if "callbacks" in params:
            self.callbacks = params.pop("callbacks")

        # Update class
        self.num_boost_round = params.pop("num_boost_round")
        self.params = params
        return self

    def _get_callbacks(self) -> list:
        # Output
        callbacks = []

        # Iterate through callbacks
        for callback in self.callbacks:
            # todo implement various callbacks - recognize string and add function
            raise NotImplementedError("Callback not implemented.")

        # Return
        return callbacks

    def get_params(self, **args):
        params = deepcopy(self.params)
        if "deep" in args:
            return params
        if self.callbacks is not None:
            params["callbacks"] = self.callbacks
        params["num_boost_round"] = self.num_boost_round
        return params

    @staticmethod
    def eval_function(prediction, d_train):
        target = d_train.get_label()
        weight = d_train.get_weight()
        # Return mean absolute error
        return (
            "mean_absolute_error",
            mean_absolute_error(target, prediction, sample_weight=weight),
            True,
        )



#  Copyright (c) 2022 by Amplo.

import os
import time
from copy import deepcopy
from datetime import datetime
from typing import TypeVar

import joblib
import numpy as np
import pandas as pd
from sklearn import ensemble, linear_model, metrics, model_selection, svm

from amplo.classification import CatBoostClassifier, LGBMClassifier, XGBClassifier
from amplo.regression import CatBoostRegressor, LGBMRegressor, XGBRegressor
from amplo.utils.logging import logger

__all__ = ["ClassificationType", "Modeller", "ModelType", "RegressionType"]


ClassificationType = TypeVar(
    "ClassificationType",
    CatBoostClassifier,
    ensemble.BaggingClassifier,
    linear_model.RidgeClassifier,
    linear_model.LogisticRegression,
    LGBMClassifier,
    svm.SVC,
    XGBClassifier,
)
RegressionType = TypeVar(
    "RegressionType",
    CatBoostRegressor,
    ensemble.BaggingRegressor,
    linear_model.LinearRegression,
    LGBMRegressor,
    svm.SVR,
    XGBRegressor,
)
ModelType = TypeVar(
    "ModelType",
    CatBoostClassifier,
    CatBoostRegressor,
    ensemble.BaggingClassifier,
    ensemble.BaggingRegressor,
    linear_model.LinearRegression,
    linear_model.LogisticRegression,
    linear_model.RidgeClassifier,
    LGBMClassifier,
    LGBMRegressor,
    svm.SVC,
    svm.SVR,
    XGBClassifier,
    XGBRegressor,
)


class Modeller:
    """
    Modeller for classification or regression tasks.

    Supported models:
        - linear models from ``scikit-learn``
        - random forest from ``scikit-learn``
        - bagging model from ``scikit-learn``
        - ~~gradient boosting from ``scikit-learn``~~
        - ~~histogram-based gradient boosting from ``scikit-learn``~~
        - XGBoost from DMLC
        - Catboost from Yandex
        - LightGBM from Microsoft

    Parameters
    ----------
    mode : str
        Model mode. Either `regression` or `classification`.
    shuffle : bool
        Whether to shuffle samples from training / validation.
    n_splits : int
        Number of cross-validation splits.
    objective : str
        Performance metric to optimize. Must be a valid string for
        `sklearn.metrics.SCORERS`.
    samples : int
        Hypothetical number of samples in dataset. Useful to manipulate behavior
        of `return_models()` function.
    needs_proba : bool
        Whether the modelling needs a probability.
    folder : str
        Folder to store models and/or results.
    dataset : str
        Name of feature set. For documentation purposes only.
    store_models : bool
        Whether to store the trained models. If true, `folder` must be
        specified to take effect.
    store_results : bool
        Whether to store the results. If true, `folder` must be specified.

    See Also
    --------
    [Sklearn scorers](https://scikit-learn.org/stable/modules/model_evaluation.html
    """

    def __init__(
        self,
        mode="regression",
        cv=None,
        objective="accuracy",
        samples=None,
        needs_proba=True,
        folder="",
        dataset="set_0",
        store_models=False,
        store_results=True,
    ):
        # Test
        assert mode in ["classification", "regression"], "Unsupported mode"
        assert isinstance(objective, str)
        assert (
            objective in metrics.SCORERS.keys()
        ), "Pick scorer from sklearn.metrics.SCORERS: \n{}".format(
            list(metrics.SCORERS.keys())
        )
        assert isinstance(samples, int) or samples is None
        assert isinstance(folder, str)
        assert isinstance(dataset, str)
        assert isinstance(store_models, bool)
        assert isinstance(store_results, bool)

        # Parameters
        self.cv = cv
        self.objective = objective
        self.scoring = metrics.SCORERS[objective]
        self.mode = mode
        self.samples = samples
        self.dataset = str(dataset)
        self.store_results = store_results
        self.store_models = store_models
        self.results = pd.DataFrame(
            columns=[
                "date",
                "model",
                "dataset",
                "params",
                "mean_objective",
                "std_objective",
                "mean_time",
                "std_time",
            ]
        )

        # Update CV if not provided
        if self.cv is None:
            if self.mode == "classification":
                self.cv = model_selection.StratifiedKFold(n_splits=3)
            elif self.mode == "regression":
                self.cv = model_selection.KFold(n_splits=3)

        # Folder
        self.folder = folder if len(folder) == 0 or folder[-1] == "/" else folder + "/"
        if (store_results or store_models) and self.folder != "":
            if not os.path.exists(self.folder):
                os.makedirs(self.folder)

        self.needs_proba = needs_proba

    def fit(self, x, y):
        # Copy number of samples
        self.samples = len(y)

        return self._fit(x, y)

    def return_models(self):
        """
        Get all models that are considered appropriate for training.

        Returns
        -------
        list of ModelType
            Models that apply for given dataset size and mode.
        """
        models = []

        # All classifiers
        if self.mode == "classification":
            # The thorough ones
            if self.samples < 25000:
                models.append(svm.SVC(kernel="rbf", probability=self.needs_proba))
                models.append(ensemble.BaggingClassifier())
                # models.append(ensemble.GradientBoostingClassifier()) == XG Boost
                models.append(XGBClassifier())

            # The efficient ones
            else:
                # models.append(ensemble.HistGradientBoostingClassifier()) == LGBM
                models.append(LGBMClassifier())

            # And the multifaceted ones
            if not self.needs_proba:
                models.append(linear_model.RidgeClassifier())
            else:
                models.append(linear_model.LogisticRegression())
            models.append(CatBoostClassifier())
            models.append(ensemble.RandomForestClassifier())

        elif self.mode == "regression":
            # The thorough ones
            if self.samples < 25000:
                models.append(svm.SVR(kernel="rbf"))
                models.append(ensemble.BaggingRegressor())
                # models.append(ensemble.GradientBoostingRegressor()) == XG Boost
                models.append(XGBRegressor())

            # The efficient ones
            else:
                # models.append(ensemble.HistGradientBoostingRegressor()) == LGBM
                models.append(LGBMRegressor())

            # And the multifaceted ones
            models.append(linear_model.LinearRegression())
            models.append(CatBoostRegressor())
            models.append(ensemble.RandomForestRegressor())

        return models

    def _fit(self, x, y):
        # Convert to NumPy
        x = np.array(x)
        y = np.array(y).ravel()

        if self.store_results and "Initial_Models.csv" in os.listdir(self.folder):
            self.results = pd.read_csv(self.folder + "Initial_Models.csv")
            for i in range(len(self.results)):
                self.print_results(self.results.iloc[i])

        else:

            # Models
            self.models = self.return_models()

            # Loop through models
            for master_model in self.models:

                # Time & loops through Cross-Validation
                val_score = []
                train_score = []
                train_time = []
                for t, v in self.cv.split(x, y):
                    t_start = time.time()
                    xt, xv, yt, yv = x[t], x[v], y[t], y[v]
                    model = deepcopy(master_model)
                    model.fit(xt, yt)
                    val_score.append(self.scoring(model, xv, yv))
                    train_score.append(self.scoring(model, xt, yt))
                    train_time.append(time.time() - t_start)

                # Append results
                result = {
                    "date": datetime.today().strftime("%d %b %y"),
                    "model": type(model).__name__,
                    "dataset": self.dataset,
                    "params": model.get_params(),
                    "mean_objective": np.mean(val_score),
                    "std_objective": np.std(val_score),
                    "worst_case": np.mean(val_score) - np.std(val_score),
                    "mean_time": np.mean(train_time),
                    "std_time": np.std(train_time),
                }
                self.results = pd.concat(
                    [self.results, pd.Series(result).to_frame().T], ignore_index=True
                )
                self.print_results(result)

                # Store model
                if self.store_models:
                    joblib.dump(
                        model,
                        self.folder
                        + type(model).__name__
                        + "_{:.4f}.joblib".format(np.mean(val_score)),
                    )

            # Store CSV
            if self.store_results:
                self.results.to_csv(self.folder + "Initial_Models.csv")

        # Return results
        return self.results

    def print_results(self, result):
        logger.info(
            "{} {}: {} \u00B1 {},    training time: {:.1f} s".format(
                result["model"].ljust(30),
                self.objective,
                f"{result['mean_objective']:.4f}",
                f"{result['std_objective']:.4f}",
                result["mean_time"],
            )
        )



#  Copyright (c) 2022 by Amplo.

"""
Base class repository of Amplo.
"""

from amplo.base.exceptions import NotFittedError
from amplo.base.objects import BaseEstimator, BaseObject, BaseTransformer, LoggingMixin



#  Copyright (c) 2022 by Amplo.

from copy import deepcopy

import numpy as np
import pandas as pd
import xgboost as xgb
from sklearn.model_selection import train_test_split


class XGBRegressor:
    _estimator_type = "regressor"
    default_params = {"verbosity": 0, "num_boost_round": 100}

    def __init__(self, **params):
        """
        XG Boost wrapper
        @param params: Model parameters
        """
        self.num_boost_round = None
        self.params = None
        self.set_params(**params)
        self.model = None
        self.callbacks = None
        self.trained = False
        self.binary = False

    @staticmethod
    def convert_to_d_matrix(x, y=None):
        # Convert input
        assert type(x) in [
            pd.DataFrame,
            pd.Series,
            np.ndarray,
        ], "Unsupported data input format"
        if isinstance(x, np.ndarray) and len(x.shape) == 0:
            x = x.reshape((-1, 1))

        if y is None:
            return xgb.DMatrix(x)

        else:
            assert type(y) in [pd.Series, np.ndarray], "Unsupported data label format"
            return xgb.DMatrix(x, label=y)

    def fit(self, x, y):
        # Split & Convert data
        train_x, test_x, train_y, test_y = train_test_split(x, y, test_size=0.1)
        d_train = self.convert_to_d_matrix(train_x, train_y)
        d_test = self.convert_to_d_matrix(test_x, test_y)

        # Model training
        self.model = xgb.train(
            self.params,
            d_train,
            evals=[(d_test, "validation"), (d_train, "train")],
            verbose_eval=False,
            num_boost_round=self.num_boost_round,
            callbacks=[self.callbacks] if self.callbacks is not None else None,
            early_stopping_rounds=100,
        )
        self.trained = True

    def predict(self, x, *args, **kwargs):
        # todo check input data
        assert self.trained is True, "Model not yet trained"
        d_predict = self.convert_to_d_matrix(x)
        return self.model.predict(d_predict, *args, **kwargs)

    def set_params(self, **params):
        # Add default
        for k, v in self.default_params.items():
            if k not in params:
                params[k] = v

        # Remove fit options
        if "callbacks" in params:
            self.callbacks = params.pop("callbacks")

        # Update class
        self.num_boost_round = params.pop("num_boost_round")
        self.params = params
        return self

    def get_params(self, **args):
        params = deepcopy(self.params)
        if "deep" in args:
            return params
        if self.callbacks is not None:
            params["callbacks"] = self.callbacks
        return params
